/* $Workfile: OsaLinux.cpp $   $Revision: 62 $    $Modtime: 1/07/10 3:28p $  */
/*******************************************************************************

  Odin TeleSystems Inc.

  OTX Telecom Adapter Driver

  @doc internal

  @module

  @normal Copyright (C) Odin TeleSystems Inc., 1998-2003, All Rights Reserved

*******************************************************************************/

#include "../Osa.h"
#include "../OsaImplementation.h"
#include "../OsaRtl.h"
#include "../OsaMemDbg.h"
#include "Lib/OsaIoctl.h"
#include "Lib/OtxHwIds.h"
#include "OsaLinuxEntry.h"
#include "OsaLinuxCppRtl.h"
#include "pcmcia/modules/otxdrv_cs.h"   // Header file shared with the OTX PCMCIA Client driver

#ifdef DAVINCI
#include "Sys/DeviceDaVinci/OdDaVinciNicDef.h"
#include "Sys/DeviceDaVinci/OdDaVinciCsiNicDef.h"
#include "OsaLinuxDaVinciI2c.h"
#include "OsaLinuxDmp.h"
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)

#define new ident_new
#define delete ident_delete
#define class ident_class
#define virtual ident_virtual			  
#define namespace ident_namespace
#define private ident_private
#define typename ident_typename

#ifndef CPP_ASMLINKAGE
#define CPP_ASMLINKAGE 
#else
#undef CPP_ASMLINKAGE
#define CPP_ASMLINKAGE 
#endif
#ifndef asmlinkage
#define asmlinkage CPP_ASMLINKAGE
#else
#undef asmlinkage
#define asmlinkage CPP_ASMLINKAGE
#endif

#ifdef DAVINCI
// GCC only supports this in "C" code not CPP, return true if types are the same
#define __builtin_types_compatible_p(x, y) \
({\
1;\
})
#endif



#undef new
#undef delete
#undef class
#undef virtual		
#undef namespace	
#undef private
#undef typename

#define ENABLE_WORKQUEUE_DBG 0 // 1=enable 0=not enabled
#else

#define ENABLE_WORKQUEUE_DBG 0 // 1=enable 0=not enabled

#endif

extern unsigned int nfnOsaLinuxCsIoControl;

static OsaLinuxCsIoControlFnP g_pfnOsaLinuxCsIoControl = NULL;


/******************************************************************************
******************************************************************************/
enum OsaLinuxDeviceTypesE {
  OSA_LINUX_DEVICE_EXTERNAL = 0,
  OSA_LINUX_DEVICE_OS_EXT,
  OSA_LINUX_DEVICE_USER_EVENT
};

/******************************************************************************
******************************************************************************/
struct OsaLinuxBaseDevice {
  OsaLinuxBaseDevice()
    : m_pProcEntry(NULL)
    , m_eLinuxDeviceType(OSA_LINUX_DEVICE_EXTERNAL)
  {}

  proc_dir_entry_otx *m_pProcEntry;
  OsaLinuxDeviceTypesE m_eLinuxDeviceType;   // e.g. an OsaUserEventDevice
};


/******************************************************************************
  Note the similarity with the regular OsaEventImpl (OsaEvent), however
  this type of event inherits from the OsaLinuxBaseDevice and also features
  the PollWait member function so that it can be made accessible from the
  poll method (like an old select() method, almost)
******************************************************************************/
struct OsaLinuxUserEventDevice: public OsaLinuxBaseDevice {
  OsaLinuxUserEventDevice()
    : m_pWaitQHead(NULL)
    , m_nEventSet(0)
    , m_pClassValidityObject(&ms_classValidityObject)
  {
    OsaLinuxUserEventDevice_Ctor(&m_pWaitQHead);
  }

  ~OsaLinuxUserEventDevice()
  {
    SetEvent();
    OsaLinuxUserEventDevice_Dtor(&m_pWaitQHead);
    m_pClassValidityObject = NULL;
  }

  bool SetEvent()
  {
    int nEventSetBefore;

    OsaSingleLock localLock(&m_accessLock);
    nEventSetBefore = m_nEventSet;
    m_nEventSet = 1;
    OsaLinuxUserEventDevice_wake_up(&m_pWaitQHead);
    return nEventSetBefore ? true : false;
  }

  bool IsSet()
  {
    OsaSingleLock localLock(&m_accessLock);
    return m_nEventSet ? true : false;
  }

  void PollWait(struct file *pFile, struct poll_table_struct *pPollTable)
  {
    osalinux_poll_wait(pFile, &m_pWaitQHead, pPollTable);
  }

  void ClearEvent()
  {
    OsaSingleLock localLock(&m_accessLock);
    m_nEventSet = 0;
  }

  static int        ms_classValidityObject;
  void              *m_pClassValidityObject;

private:
  void              *m_pWaitQHead;
  OsaUint32         m_nEventSet;
  OsaSpinLock       m_accessLock;
};


/******************************************************************************
******************************************************************************/
struct OsaLinuxDevice: public OsaLinuxBaseDevice, public OsaDevice {
  OsaLinuxDevice()
    : m_pPciDev(NULL)
    , m_bInterruptConnected(false)
    , m_pDeviceCompanionDriver(NULL)
    , m_sTasklets1(NULL)
  {}
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
public:

  /******************************************************************************
    OsaLinuxAddToWorkQueue0
    Work queue function to add this object's work to the queue. 
    Need a unique function for each. The work queue will only execute
    one if two IRQ's are called before the queue can start the first one.
  ******************************************************************************/
  static void OsaLinuxAddToWorkQueue0(IN unsigned long queueData)
  {
    if(queueData == 0) {
      OSA_DPRINTF(("Nothing to do!! queueData is 0x%x \n", 
                  queueData));
      return;
    }
    
    // Part of OsaLinuxEntry.c to finish the taskqueue startup and perform
    // cleanup as needed for the taskqueue. 
    OsaLinuxIsrWorkQueue(queueData);
  }
  /******************************************************************************
    OsaLinuxAddToWorkQueue1
  ******************************************************************************/
  static void OsaLinuxAddToWorkQueue1(IN unsigned long queueData)
  {
    if(queueData == 0) {
      OSA_DPRINTF(("Nothing to do!! queueData is 0x%x \n", 
                  queueData));
      return;
    }
    
    // Part of OsaLinuxEntry.c to finish the taskqueue startup and perform
    // cleanup as needed for the taskqueue. 
    OsaLinuxIsrWorkQueue(queueData);
  }
  /******************************************************************************
    OsaLinuxAddToWorkQueue2
  ******************************************************************************/
  static void OsaLinuxAddToWorkQueue2(IN unsigned long queueData)
  {
    if(queueData == 0) {
      OSA_DPRINTF(("Nothing to do!! queueData is 0x%x \n", 
                  queueData));
      return;
    }
    
    // Part of OsaLinuxEntry.c to finish the taskqueue startup and perform
    // cleanup as needed for the taskqueue. 
    OsaLinuxIsrWorkQueue(queueData);
  }
  /******************************************************************************
    OsaLinuxSetWorkQueuePtr
    Helper function to setup the workqueue pointer.
  ******************************************************************************/
  void OsaLinuxSetWorkQueuePtr(void)
  {
    int queueOk;

    m_nIsrIndex++;
    if(m_nIsrIndex >= 3) {
      m_nIsrIndex = 0;
    }

    // Set the workqueue function for this non-critical interrupt
    switch (m_nIsrIndex) {
      case 0:  
        m_pWorkqueueFunc = (void *) &(this->OsaLinuxAddToWorkQueue0); 
        break;      
      case 1:  
        m_pWorkqueueFunc = (void *) &(this->OsaLinuxAddToWorkQueue1); 
        break;      
      case 2:  
        m_pWorkqueueFunc = (void *) &(this->OsaLinuxAddToWorkQueue2); 
        break;      

      default:
        // This should never happen
        m_nIsrIndex = 0; // try and get it back in sync.
        m_pWorkqueueFunc = (void *) &(this->OsaLinuxAddToWorkQueue0);
        // This could cause an IRQ to be lost (not processed).
        OSA_EPRINTF(("OsaLinux: SoftIrq start function synchronization error.\n"));
    }

    // Pass a OsaLinuxDevice object pointer to the work queue
    queueOk = OsaLinuxAddWorkQueue((void *)&m_sTasklets1, 
                                   (void *)&(m_nWorkQueueData),
                                   (void *)&(*m_pWorkqueueFunc));
    if(queueOk != 0) {
      OSA_EPRINTF(("OsaLinux: Not able to add task to softIsr\n"));
    }


    return;
  }
#endif
  void   *m_pPciDev;
  bool    m_bInterruptConnected;
  void   *m_pDeviceCompanionDriver;  // A pointer to the implementation of this device in
                                     // companion driver (e.g.the PCMCLA client driver)

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
  unsigned long volatile m_nWorkQueueData;
  void   *m_nPendingInterrupts;
  bool    m_bSoftIsrActive;
  void * volatile m_sTasklets1;
  void (*m_pWorkqueueFunc)(IN unsigned long);
  // Each card needs a unique workqueue starting task,
  // this index keep track of the last one used so each interrupt
  // will get a unique funciton. (If 2 interrupt functions are sent at the 
  // same time the workqueue only starts one of them.)
  OsaUint32 m_nIsrIndex;  
#endif
};


/******************************************************************************
******************************************************************************/
static void OsaLinuxInit();

static int OsaLinuxEnumAndAddAndStartAllPciBoards();
//static int OsaLinuxStartFirstPcmciaBoard();

static bool OsaLinuxAddDevice(
  IN OsaBusTypeE eBusType,
  IN OsaUint16 nVid,
  IN OsaUint16 nDid,
  IN OsaUint16 nSid,
  IN wchar_t *pwcszVendorString,
  OUT OsaLinuxDevice **ppExternalDevice
  );

static bool OsaLinuxExtractPciResources(
  IN void *pPciDev,
  IN OsaResourcesS **ppResources
  );

static void OsaLinuxReturnResources(
  IN OsaLinuxDevice *pExternalDevice
  );


static bool OsaLinuxStartDevice(
  IN OsaLinuxDevice *pExternalDevice,
  IN OsaResourcesS *pResourceList
  );

static bool OsaLinuxRemoveDevice(
  IN OsaLinuxDevice *pExternalDevice
  );


#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
static void OsaLinuxWorkQueueIsr(IN void *pContext);
#endif


static int ms_coEventsCreated = 0;
int OsaLinuxUserEventDevice::ms_classValidityObject;

/******************************************************************************
  Entry function exported to the pure "C" init_module()
******************************************************************************/
int OsaLinuxDriverEntry(void)
{
  int rc;
  OSA_DPRINTF(("OsaLinux: Creating OSA \n"));
  // Create Linux specific OSA

  OsaLinuxInit();

  // The drivers main entry point, must be exported as extern "C"
  // Sets up the other entry points in the form of pointers, very much like
  // OdOsa does for the operating system. This function is the only
  // function the driver must export so that Osa can bootstrap the driver into action.
  OsaString dummy;
#if 1 // OsaString does not work properly on empty strings, so we assign a dummy value to it /HJ 2000-10-18
  dummy.Append("Dummy");
#endif


  if (OSA_SUCCESS(::OsaDriverEntry(dummy.GetWcString()))) {

    if (!OsaCheckCallbacks()) {
      OSA_EPRINTF(("OsaLinux: The OSA base set of callbacks were not hooked up\n"));
      return -1;
    }

    // Enumerate OTX PCI boards
    // Note: The OTX PCMCIA boards are not enumerated. The OTX PCMCIA Client driver will
    // call OsaLinuxAddDeviceCallback() when an OTX PCMCIA board is inserted.
    rc = OsaLinuxEnumAndAddAndStartAllPciBoards();

    if (rc) {
      OSA_DPRINTF(("OsaLinux: Calling cleanup_module before failing the loading, do we need to do this?\n"));
      cleanup_module(); // Must find out if driver loads or not when returning an error code
    } else {
      // Create a bootstrap device for OS extensions (name to be known by user mode)
      OsaLinuxBaseDevice *pOsExt = new OsaLinuxBaseDevice;
      pOsExt->m_pProcEntry = OsaLinuxCreateProcDevice(L"osaext", pOsExt);
      pOsExt->m_eLinuxDeviceType = OSA_LINUX_DEVICE_OS_EXT;
    }

    return rc;
  } else {
    return -1;
  }

  return 0;
}


#ifdef DAVINCI

/******************************************************************************
  Entry function exported to the pure "C" init_module()
******************************************************************************/
extern int OsaLinuxDaVinciDriverEntry(void)
{
  int rc = 1;
  OsaUint32 nReadVal;
  OsaUint32 nDmp;
  

  OsaBusTypeE  eBusType = OSA_BUS_LOCAL;
  OsaUint16 nVid = OTX_ODIN_VENDOR_ID;
  OsaUint16 nDid;
  OsaUint16 nSid; 

  // When OsaLinuxAddDevice() was called it is set to NULL, do the same here
  // this function is not called. OsaLinuxAddDevice is normally used in PCI bus
  wchar_t *pwcszVendorString = NULL; 

  OsaLinuxDevice *pExternalDevice = NULL;
  OsaString strExternalDeviceName;
  OsaInternalDeviceRep internalDeviceRep = NULL;


  OSA_DPRINTF(("OsaLinux: Creating OSA \n"));


  OsaLinuxDmpReadReg32(OD_DMP_DIR01_REG, &nReadVal);
  OSA_DPRINTF(("OsaLinux: GPIO DIR01 = 0x%.8X\n", nReadVal));

  // Some code in the kernel sets GPIO2 to an output. Here we force it to an input.
  OsaLinuxDmpWriteReg32(OD_DMP_DIR01_REG, nReadVal | 0x00000004);

  OsaLinuxDmpReadReg32(OD_DMP_IN_DATA01_REG, &nReadVal);
  OSA_DPRINTF(("OsaLinux: GPIO 31-0 = 0x%.8X\n", nReadVal));

#if 0 // Temporary to simulate an ALVIS CSI
  nReadVal |= OD_DMP_GPIO2;
#endif

  if (nReadVal & OD_DMP_GPIO2) {  // GPI02 is the Platform ID bit (CSI or ASM/PCIe)
    OSA_DPRINTF(("OsaLinux: DMP host is CSI. Reading Did and Sid from Eeprom.\n"));
    if (!(nReadVal & OD_DMP_GPIO6)) {
      // SLOT_ID0: GPIO6 = 0 means that this is DMP0 of the DMM
      
      OSA_DPRINTF(("OsaLinux: DMP0. Reading Did and Sid from Eeprom.\n"));
      // Read nDid (Device ID)
      {
        OsaInt32 rc;
        
        // Dummy read. For some reason the first read from the Eeprom reads from address 0 even though we provide a different address
        rc = OsaLinuxDaVinciEepromReadBuf(OD_NICDV_CSI_EEPROM_DEV_ID_OFFSET, (OsaUint8 *)&nDid, 2);
        
        // The real read...
        rc = OsaLinuxDaVinciEepromReadBuf(OD_NICDV_CSI_EEPROM_DEV_ID_OFFSET, (OsaUint8 *)&nDid, 2);
        
        if (!rc) {
          if (nDid == 0xFFFF) {
            nDid = OTX_PROD_ID_DMP_ALVIS_CSI_UNCONFIGURED;
            OSA_DPRINTF(("Board Device ID has not been programmed. Forcing it to 0x%.4X (OTX_PROD_ID_DMP_ALVIS_CSI_UNCONFIGURED)\n", nDid));
          } else {
            OSA_DPRINTF(("Board Device ID = 0x%.4X\n", nDid));
          }
        } else {
          OSA_EPRINTF(("OsaLinuxDaVinciDriverEntry: Could read DeviceId from EEPROM\n"));
          return false;
        }
      }
      
      // Read nSid (SubSystem ID, which is the product version in OTX terminology)
      {
        OsaInt32 rc;
        
        rc = OsaLinuxDaVinciEepromReadBuf(OD_NICDV_CSI_EEPROM_SUBSYSTEM_ID_OFFSET, (OsaUint8 *)&nSid, 2);
        
        if (!rc) {
          if (nSid == 0xFFFF) {
            OSA_DPRINTF(("SubSystem ID has not been programmed (0xFFFF)\n"));
          } else {
            OSA_DPRINTF(("SubSystem ID = 0x%.4X\n", nSid));
          }
        } else {
          OSA_EPRINTF(("OsaLinuxDaVinciDriverEntry: Could read SubSystemID from EEPROM\n"));
          return false;
        }
      }
    } else {
      // SLOT_ID0: GPIO6 = 1 means that this is DMP1 of the DMM
      OSA_DPRINTF(("OsaLinux: DMP1. \"Co-processor\" without access to peripherals. nDid = 0x%.4X, nSid = 0x%.4X\n", OTX_PROD_ID_DMP_CSI_SECONDARY, OD_NIC_SUBSYSTEM_ID_REV_1_0));
      nDid = OTX_PROD_ID_DMP_CSI_SECONDARY;
      nSid = OD_NIC_SUBSYSTEM_ID_REV_1_0;
    }

  } else {
    nDid = OTX_PROD_ID_DMP_ASM_PCIE; 
    nSid = OD_NIC_SUBSYSTEM_ID_REV_1_0;
    OSA_DPRINTF(("OsaLinux: DMP host is ASM/PCIe (DeviceId=0x%.4X, SubSystemID=0x%.4X)\n", nDid, nSid));
    nDmp =  ((nReadVal & OD_DMP_GPIO3) ? 4 : 0) + ((nReadVal & OD_DMP_GPIO5) ? 2 : 0) + ((nReadVal & OD_DMP_GPIO6) ? 1 : 0);
    OSA_DPRINTF(("OsaLinux: DMP#%d\n", nDmp));
  }

  // Create Linux specific OSA

  OsaLinuxInit();

  // The drivers main entry point, must be exported as extern "C"
  // Sets up the other entry points in the form of pointers, very much like
  // OdOsa does for the operating system. This function is the only function 
  // the driver must export so that Osa can bootstrap the driver into action.
  OsaString dummy;
  // OsaString does not work properly on empty strings, 
  // so we assign a dummy value to it /HJ 2000-10-18
  dummy.Append("Dummy");


  if (OSA_SUCCESS(::OsaDriverEntry(dummy.GetWcString()))) {

    if (!OsaCheckCallbacks()) {
      OSA_EPRINTF(("OsaLinuxDaVinciDriverEntry: OSA base callbacks not defined\n"));
      return false;
    }

    if (!OSA_SUCCESS((*g_sOsaDriverEntryPoints.pfnAddBoard)(eBusType, 
                                                            nVid, 
                                                            nDid, 
                                                            nSid, 
                                                            pwcszVendorString, // set to null, not used
                                                            &internalDeviceRep, 
                                                            &strExternalDeviceName))) {
      OSA_EPRINTF(("OsaLinuxDaVinciDriverEntry: OSA Add board failed\n"));
      return false;
    }

    if((&strExternalDeviceName == NULL) ||
       (&internalDeviceRep == NULL)) {
      OSA_EPRINTF(("OsaLinuxDaVinciDriverEntry: OSA Add board returned null pointer\n"));
    }

    // There should be a device for use to use. No need to enumerate.
    if (!OSA_SUCCESS(OsaCreateExternalDevice(
      strExternalDeviceName,                // Externally known name of the board device
      internalDeviceRep,                    // A reference by which the device is known internally (Driver)
      (void **) &pExternalDevice            // A reference by which the device is known externally (OSA and System)
      )))
    {
      // Cleanup after (*AddBoardFnP)() since it fails the caller won't call 
      // OsaLinuxRemoveDevice (caller thinks device was never added)
      OSA_EPRINTF(("OsaLinux: Failed to create external device\n", eBusType));
      (*g_sOsaDriverEntryPoints.pfnRemoveBoard)(internalDeviceRep);
      return false;
    }

    OSA_ASSERT(pExternalDevice);
    pExternalDevice->m_bBoard = true;

    // Create a resource list for the claimed and mapped resources.
    OsaResourcesS *pClaimedResources = new OsaResourcesS;

    OSA_ASSERT(pClaimedResources);

    if (!pClaimedResources) {
      OSA_DPRINTF(("OsaLinuxDaVinciDriverEntry: Out of memory OTX DaVinci device\n"));
      return false;
    }

    // ********************************************************************
    // We must attach the resources to the device so they get released with 
    // the device. We will soon start claiming resources and may return with 
    // an error. Any error returned by this function will result in a call to 
    // OsaLinuxRemoveDevice and OsaLinuxReturnResources. Return resources will 
    // return any resources we have claimed and mapped here, but only
    // if they can be found through pExternalDevice->m_pResources
    // ********************************************************************

    // We have no resources claimed yet, but an empty OsaResourcesS object is OK
    pExternalDevice->m_pResources = pClaimedResources;  

    pExternalDevice->m_pResources->coPortAreas   = 0;
    pExternalDevice->m_pResources->coMemoryAreas = 0;
    // The DaVinci driver handles the interrupt
    // connection/disconnection.
    pExternalDevice->m_bInterruptConnected = true;   
    // used to show that the soft ISR is active
    pExternalDevice->m_bSoftIsrActive = false; 

    // Needed since we perform a check on this variable in OdPnpDevice
    // Setting this variable to true will tell the PnPDevice to
    // connect the Notification Callback routine
    // The interrupt is connected in ConnectNotificationCallbacks()
    OSA_DPRINTF(("OsaLinuxDaVinciDriverEntry: Using interrupt level 56 (Bank0) for interrupts to the board class\n"));
    pExternalDevice->m_pResources->sInterrupt.bHaveInterrupt = true;
    pExternalDevice->m_pResources->sInterrupt.nLevel = 56; 

    // Start DaVinci using the claimed and mapped resources
    if (!OSA_SUCCESS((*g_sOsaDriverEntryPoints.pfnStartBoard)(pExternalDevice->m_internalDeviceRep, 
                                                              pExternalDevice, 
                                                              pExternalDevice->m_pResources))) {

      // If the device could not be started we must delete it
      // Any resources claimed will be released by OsaLinuxRemoveDevice 
      // calling OsaLinuxReturnResources
      OSA_DPRINTF(("OsaLinuxDaVinciDriverEntry: Failed to start OTX DaVinci device\n"));
      OsaLinuxRemoveDevice(pExternalDevice);

      return false;
    }

    OSA_DPRINTF(("OsaLinuxDaVinciDriverEntry: Started device\n"));

    // Create a bootstrap device for OS extensions (name to be known by user mode)
    OsaLinuxBaseDevice *pOsExt = new OsaLinuxBaseDevice;
    pOsExt->m_pProcEntry = OsaLinuxCreateProcDevice(L"osaext", pOsExt);
    pOsExt->m_eLinuxDeviceType = OSA_LINUX_DEVICE_OS_EXT;
  
    OSA_DPRINTF(("OsaLinuxDaVinciDriverEntry: Driver loaded correctly."));

    return rc;
  } else {
    return false;
  }
}
#else
extern int OsaLinuxDaVinciDriverEntry(void)
{
  OSA_DPRINTF(("OsaLinuxDaVinciDriverEntry: Error DaVinci build not defined\n"));

  return false;
}
#endif

/******************************************************************************
  Entry function exported to the pure "C" cleanup_module()
******************************************************************************/
void OsaLinuxUnloadEntry()
{
  if (g_sOsaDriverEntryPoints.pfnUnload) {
    (*g_sOsaDriverEntryPoints.pfnUnload)();
  }

#if DBG
  if (OsaMemoryGetCurrentUsage()) {
    OSA_EPRINTF(("OsaLinux: Memory blocks no freed:\n"));
    OsaPrintCurrentlyUsed();
    OSA_EPRINTF(("OsaLinux: Dynamic memory leak is %d bytes\n", OsaMemoryGetCurrentUsage()));
  } else {
    OSA_DPRINTF(("OsaLinux: All dynamic memory freed properly\n"));
  }
#endif

  return;
}


/******************************************************************************
******************************************************************************/
void OsaLinuxInit()
{
  OsaInit();
  OsaInitLinuxRtl();
}


/******************************************************************************

  Called by the driver whenever the appearance of a named device must be
  presented to user mode apps or other drivers or the system or a combination
  of all. In particular, all board devices *must* be defined for the system, with
  a visible name, hence this function must be called from the drivers AddBoard
  function at the least.

  Od must supply a reference by which the device will be known in Od.

  The function returns a reference by which the device is known in Osa.

******************************************************************************/
OsaResultE OsaCreateExternalDevice(
  IN const OsaString &strDeviceName,            // The device name we want to present externally
  IN OsaInternalDeviceRep internalDeviceRep,       // A reference by which the device is known internally
  OUT OsaExternalDeviceRep *pExternalDeviceRep
)
{
  OsaLinuxDevice *pExternalDevice = new OsaLinuxDevice;
  OSA_ASSERT(pExternalDevice);
  if (!pExternalDevice) return OSA_RES_FAILED;

  OSA_DPRINTF(("OsaLinux: Created external device 0x%p\n", pExternalDevice));

  *pExternalDeviceRep = NULL;

  pExternalDevice->m_pProcEntry = OsaLinuxCreateProcDevice(strDeviceName.GetWcString(), pExternalDevice);
  pExternalDevice->m_eLinuxDeviceType = OSA_LINUX_DEVICE_EXTERNAL;

  if (!pExternalDevice->m_pProcEntry) return OSA_RES_FAILED;

  // Save the name, chosen by the internal part the driver, in the device extension
  pExternalDevice->m_strName = strDeviceName;

  pExternalDevice->m_internalDeviceRep = internalDeviceRep;

  *pExternalDeviceRep =  pExternalDevice;

  return OSA_RES_SUCCESS;
}


/******************************************************************************
******************************************************************************/
void OsaDestroyExternalDevice(
  IN OsaExternalDeviceRep externalDeviceRep       // A reference by which the device is known internally
)
{
  OsaLinuxDevice *pExternalDevice = reinterpret_cast<OsaLinuxDevice *>(externalDeviceRep);

  OSA_DPRINTF(("OsaLinux: Destroying external device 0x%p\n", pExternalDevice));
  OsaLinuxRemoveProcDevice(pExternalDevice->m_pProcEntry);
  pExternalDevice->m_pProcEntry = NULL;
  if(pExternalDevice->m_sTasklets1 != NULL) osalinux_kfree(pExternalDevice->m_sTasklets1);
  delete pExternalDevice;
}



/******************************************************************************
******************************************************************************/
bool OsaDoBoardSyncCall(IN OsaExternalDeviceRep pBoard, IN OsaSyncCallFnP pfnSyncCall, IN OUT void *pContext)
{
  OsaLinuxDevice *pExternalDevice = reinterpret_cast<OsaLinuxDevice *>(pBoard);

  OSA_ASSERT(pExternalDevice && pExternalDevice->m_bBoard && pfnSyncCall);

  if (!pExternalDevice || !pExternalDevice->m_bBoard || !pfnSyncCall) {
    return false;
  }

  unsigned long nFlags;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
  save_flags(nFlags);
  cli();
#else
  osalinux_save_flags_cli(&nFlags);
#endif
  bool bResult = (*pfnSyncCall)(pExternalDevice->m_internalDeviceRep, pContext);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
  restore_flags(nFlags);
#else
  osalinux_restore_flags(nFlags);
#endif
  return bResult;
}

/******************************************************************************
******************************************************************************/
void OsaRequestDeferredBoardNotification(IN OsaExternalDeviceRep pBoard, OsaIrpP pIrp, void *pContext)
{
#if ENABLE_WORKQUEUE_DBG == 1
  static unsigned long ntstPtr = 1;
#endif
  OsaLinuxDevice *pExternalDevice = reinterpret_cast<OsaLinuxDevice *>(pBoard);

  OSA_ASSERT(pExternalDevice && pExternalDevice->m_bBoard);


#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
#if 0 // Enable here to use Tasklet to process the reset of the interrupt in a task (DPC)
  {
    if(pExternalDevice->m_bSoftIsrActive == true) {
      // The last IRQ enabled NC-interrupts but is still not to
      // the end of OsaLinuxWorkQueueIsr(). May need to put a
      // thread sleep at the start of the function so the first
      // one can finish. Can not sleep here this is an interrupt.
#if ENABLE_WORKQUEUE_DBG == 1
      OSA_DPRINTF(("OsaLinux: Soft ISR is active\n"));
#endif
    }
    void *workQueue = reinterpret_cast<OsaLinuxDevice *>(pBoard);
    // The work queue will need to know what interrupts were set
    pExternalDevice->m_nPendingInterrupts = pContext;
    // The tasklet needs an unsigned long that contains the data. If
    // we passed the object directly it would need a unsigned long to
    // store the object pointer. But this is a function that is used by
    // all cards in the PC so each card need it's own variable to store
    // the object pointer.
    // Store object pointer in ULONG and pass that to the work queue.
    pExternalDevice->m_nWorkQueueData = (unsigned long) workQueue;

    // Set the IRQ start function for this interrupt.
    pExternalDevice->OsaLinuxSetWorkQueuePtr();

#if ENABLE_WORKQUEUE_DBG == 1
    // This check is only needed for debug. 
    // If the pointer which starts here, then which is passed 
    // to the workqueue and is finally returned to OsaLinuxWorkQueueIsr(),  
    // changes then two workqueues are active. This should only happen 
    // if two or more cards are in the PC.
    if(ntstPtr != (unsigned long) &(pExternalDevice->m_sTasklets1)) {
      ntstPtr = (unsigned long) &(pExternalDevice->m_sTasklets1);
      OSA_DPRINTF(("SoftIsr tstPtr-1 = 0x%x\n", ntstPtr));
    }
#endif

  }
#else // Process everything in the interrupt
  (*pExternalDevice->m_pfnDeferredBoardNotificationCallback)(pExternalDevice->m_internalDeviceRep, (OsaIrpP) pIrp, pContext);
#endif

#else  

  (*pExternalDevice->m_pfnDeferredBoardNotificationCallback)(pExternalDevice->m_internalDeviceRep, (OsaIrpP) pIrp, pContext);

#endif

  return;
}





/******************************************************************************
  A static wrapper callback.
******************************************************************************/
void OsaLinuxInterruptCallback(
  IN int nIrq,
  IN void *pContext,
  struct pt_regs *pRegs
)
{
  OsaLinuxDevice *pExternalDevice = reinterpret_cast<OsaLinuxDevice *>(pContext);
  (*pExternalDevice->m_pfnBoardNotificationCallback)(pExternalDevice->m_internalDeviceRep);
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
/******************************************************************************
  Local Work queue call back function
    This call is the start of a new "kernel thread:process:task" that was 
    started by the kernel using the work queue implementation.
******************************************************************************/
static void OsaLinuxWorkQueueIsr(IN void *pContext)
{
#if ENABLE_WORKQUEUE_DBG == 1
  static OsaUint32 workThreadCnt = 0; // No more then one thread per card
  static unsigned long ntstPtr = 3;
  static unsigned long ntstPtr2 = 0;
#endif
  OsaUint32 *tempPtr;
  OTX_INT32 pendingInterrupts;

  tempPtr = (OsaUint32 *) pContext;
//  OSA_DPRINTF(("OsaLinux: pContext = 0x%x\n", tempPtr));

  // The pointer to the OsaLinuxDevice was passed to the work queue
  // need to cast the void * to a OsaLinuxDevice *.
  OsaLinuxDevice *pExternalDevice = reinterpret_cast<OsaLinuxDevice *>(tempPtr);

#if ENABLE_WORKQUEUE_DBG == 1
  // Number of active thread at the same time
  workThreadCnt++;
  if(workThreadCnt > 1) {
    // There is at least one other active thread. This should only cause 
    // issues if the ISR is not able to have multiple threads running.
    OSA_DPRINTF(("OsaLinux: %d workqueue thread are active\n", workThreadCnt));
  }

  // This check is only needed for debug. 
  // If the pointer set in OsaRequestDeferredBoardNotification()
  // which is passed to the workqueue has changed then two tasklets  
  // are active. This should only happen if two or more cards are in the PC.
  if(ntstPtr != (unsigned long) &(pExternalDevice->m_sTasklets1)) {
    ntstPtr = (unsigned long) &(pExternalDevice->m_sTasklets1);
//    OSA_DPRINTF(("SoftIsr tstPtr-3 = 0x%x\n", ntstPtr));
  }  
#endif

  if(pContext == 0) {
    OSA_DPRINTF(("OsaLinux: pContext = 0\n"));
  }
  else
  if(pExternalDevice == NULL) {
    OSA_DPRINTF(("OsaLinux: pExternalDevice = 0\n"));
  }
  else
  if((*pExternalDevice->m_pfnBoardNotificationCallback) == NULL) {
     OSA_DPRINTF(("OsaLinux: *pExternalDevice = 0\n"));
  }
  else {
#if ENABLE_WORKQUEUE_DBG == 1
    pExternalDevice->m_bSoftIsrActive = true; // Soft ISR is running
#endif
    pendingInterrupts = pExternalDevice->m_nPendingInterrupts;
    // This will call the objects interrupt service routine
    (*pExternalDevice->m_pfnDeferredBoardNotificationCallback)(pExternalDevice->m_internalDeviceRep, NULL, pendingInterrupts);
#if ENABLE_WORKQUEUE_DBG == 1
    // Note at this point a new DPC(thread) could start.
    // The NC-interrupts have been enabled.
    pExternalDevice->m_bSoftIsrActive = false; // Soft ISR is finished
#endif
  }

  // Finished this task. 
  // Note at this point a new DPC(thread) could start.
  // The NC-interrupts have been enabled.
#if ENABLE_WORKQUEUE_DBG == 1
  workThreadCnt--;
#endif

  return;
}

/******************************************************************************
  Work queue call back function. This function is called from OsaLinuxEntry.c
  It is the entry point for the interrupt object to the card.
******************************************************************************/
void OsaLinuxWorkQueueFunction(void *pContext)
{
  OsaLinuxWorkQueueIsr(pContext);
}
#endif


/******************************************************************************
  Entry function exported to the pure "C" OsaDriverEntry.c
******************************************************************************/
int OsaLinuxOpenDeviceEntry(struct file *pFile, void *pContext)
{
  OsaResultE eResult = OSA_RES_SUCCESS;
  OsaIrp sIrp;

  if (!pFile || !pContext) {
    OSA_DPRINTF(("OsaLinux: OsaLinuxOpenDeviceEntry: pFile:%x  pContext:%x\n", pFile, pContext));
    return -1;
  }

  sIrp.nIoCode          = 0;
  sIrp.pInputBuffer     = NULL;
  sIrp.pOutputBuffer    = NULL;
  sIrp.cbInput          = 0;
  sIrp.cbOutputCapacity = 0;
  sIrp.nOwnerId         = (OsaUint32) pFile;
  sIrp.cbOutput         = 0;

  OsaLinuxBaseDevice *pBaseDevice = reinterpret_cast<OsaLinuxBaseDevice *>(pContext);

  switch (pBaseDevice->m_eLinuxDeviceType) {
  case OSA_LINUX_DEVICE_OS_EXT:
    OSA_DPRINTF(("OsaLinux: Opening OsExtension device\n"));
    break;
  case OSA_LINUX_DEVICE_USER_EVENT:
    {
      OSA_DPRINTF(("OsaLinux: Opening UserEvent device\n"));
    }
    break;
  default:
    {
      OsaLinuxDevice *pExternalDevice = reinterpret_cast<OsaLinuxDevice *>(pContext);
      OSA_DPRINTF(("OsaLinux: Opening external device\n"));
      eResult = (*g_sOsaDriverEntryPoints.pfnOpenDevice)(pExternalDevice->m_internalDeviceRep, &sIrp);
    }
    break;
  }

  // Translate to Linux return codes
  switch (eResult) {
  case OSA_RES_SUCCESS:
    //OSA_DPRINTF(("OsaLinux: pfnCloseDevice returning %d bytes\n", pHdr->cbReturned));
    return 0;
  default:
    OSA_DPRINTF(("OsaLinux: pfnCloseDevice returning error (0) bytes\n"));
    return -1;
  }
}



/******************************************************************************
  Entry function exported to the pure "C" OsaDriverEntry.c
******************************************************************************/
int OsaLinuxCloseDeviceEntry(struct file *pFile, void *pContext)
{
  OsaResultE eResult = OSA_RES_SUCCESS;
  OsaIrp sIrp;

  sIrp.nIoCode          = 0;
  sIrp.pInputBuffer     = NULL;
  sIrp.pOutputBuffer    = NULL;
  sIrp.cbInput          = 0;
  sIrp.cbOutputCapacity = 0;
  sIrp.nOwnerId         = (OsaUint32) pFile;
  sIrp.cbOutput         = 0;

  OsaLinuxBaseDevice *pBaseDevice = reinterpret_cast<OsaLinuxBaseDevice *>(pContext);

  switch (pBaseDevice->m_eLinuxDeviceType) {
  case OSA_LINUX_DEVICE_OS_EXT:
    {
      // No special device class is used but we do the reinterpret_cast anyway (for consistency)
      OsaLinuxBaseDevice *pDevice = reinterpret_cast<OsaLinuxBaseDevice *>(pBaseDevice);
      OSA_DPRINTF(("OsaLinux: Closing OsExtension device\n"));
      // The device is not deleted.
    }
    break;
  case OSA_LINUX_DEVICE_USER_EVENT:
    {
      OsaLinuxUserEventDevice *pDevice = reinterpret_cast<OsaLinuxUserEventDevice *>(pBaseDevice);
      OSA_DPRINTF(("OsaLinux: Closing User Event, proc_entry deleted\n"));
      OsaLinuxRemoveProcDevice(pDevice->m_pProcEntry);
      delete pDevice;
    }
    break;
  default:
    {
      OsaLinuxDevice *pExternalDevice = reinterpret_cast<OsaLinuxDevice *>(pContext);
      OSA_DPRINTF(("OsaLinux: Closing external device\n"));
      eResult = (*g_sOsaDriverEntryPoints.pfnCloseDevice)(pExternalDevice->m_internalDeviceRep, &sIrp);
    }
    break;
  }

  // Translate to Linux return codes
  switch (eResult) {
  case OSA_RES_SUCCESS:
    //OSA_DPRINTF(("OsaLinux: pfnCloseDevice returning %d bytes\n", pHdr->cbReturned));
    return 0;
  default:
    OSA_DPRINTF(("OsaLinux: pfnCloseDevice returning error (0) bytes\n"));
    return -1;
  }
}


/******************************************************************************
  Entry function exported to the pure "C" OsaDriverEntry.c
******************************************************************************/
int OsaLinuxIoDeviceControlEntry(void *pFile, void *pContext, OsaUint32 nIoControlCode, OsaLinuxIoCtlHdr *pHdr, void *pSystemBuffer)
{
  OsaResultE eResult = OSA_RES_SUCCESS;
  OsaIrp sIrp;

  sIrp.nIoCode          = nIoControlCode;
  sIrp.pInputBuffer     = pSystemBuffer;
  sIrp.pOutputBuffer    = pSystemBuffer;
  sIrp.cbInput          = pHdr->cbInBufSize;
  sIrp.cbOutputCapacity = pHdr->cbOutBufSize;
  sIrp.nOwnerId         = (OsaUint32) pFile;
  sIrp.cbOutput         = 0;  // If the callee wants to return something he must set this value

  OsaLinuxBaseDevice *pBaseDevice = reinterpret_cast<OsaLinuxBaseDevice *>(pContext);
  switch (pBaseDevice->m_eLinuxDeviceType) {

  case OSA_LINUX_DEVICE_OS_EXT:
    {
      // No special device class is used but we do the reinterpret_cast anyway (for consistency)
      OsaLinuxBaseDevice *pDevice = reinterpret_cast<OsaLinuxBaseDevice *>(pBaseDevice);
      switch (nIoControlCode) {

      case OSA_LINUX_OSEXT_IO_CREATE_EVENT:
        {
          // Create a proc entry and a OsaLinuxUserEventDevice. Attach the two to eachouther and return
          // the chosen name of the event device to the user lib
          OsaLinuxIoOutCreateEvent *pOut = reinterpret_cast<OsaLinuxIoOutCreateEvent *>(sIrp.pOutputBuffer);

          OsaString strName;
          strName.Append(OSA_USER_EVENT_NAME_PREFIX);
          strName.Append(ms_coEventsCreated++);

          OsaLinuxUserEventDevice *pUserEvent = new OsaLinuxUserEventDevice;
          pUserEvent->m_pProcEntry = OsaLinuxCreateProcDevice(strName.GetWcString(), pUserEvent);
          pUserEvent->m_eLinuxDeviceType = OSA_LINUX_DEVICE_USER_EVENT;

          // Return the name so that the user mode lib can open it and get a user mode handle
          if (sIrp.cbOutputCapacity - sizeof(void*) > strName.GetLength()) {
            pOut->pEventDevice = pUserEvent;
            wcsztosz(pOut->szProcEntryName, strName.GetWcString());
            sIrp.cbOutput = sizeof(OsaLinuxIoOutCreateEvent);
          } else {
            eResult = OSA_RES_FAILED;
          }
        }
        break;

      case OSA_LINUX_OSEXT_IO_DELETE_EVENT:
        {
          OsaLinuxUserEventDevice *pDevice = reinterpret_cast<OsaLinuxUserEventDevice *>(pBaseDevice);
          OsaLinuxRemoveProcDevice(pDevice->m_pProcEntry);
          delete pDevice;
        }
        break;

      default:
        break;
      }

    }
    break;

  case OSA_LINUX_DEVICE_USER_EVENT:
    {
      OsaLinuxUserEventDevice *pDevice = reinterpret_cast<OsaLinuxUserEventDevice *>(pBaseDevice);
      switch (nIoControlCode) {
      case OSA_LINUX_USREV_IO_SET_EVENT:
        {
          pDevice->SetEvent();
        }
        break;

      default:
        OSA_DPRINTF(("OsaLinux: Unknown OsaLinux Internal I/O code 0x%x", nIoControlCode));
        break;
      }
    }
    break;
  default:
    OsaLinuxDevice *pExternalDevice = reinterpret_cast<OsaLinuxDevice *>(pContext);

    eResult = (*g_sOsaDriverEntryPoints.pfnIoControlDevice)(pExternalDevice->m_internalDeviceRep, &sIrp);
    break;
  }
  // Translate to NTWDM status codes
  switch (eResult) {
  case OSA_RES_SUCCESS:
    pHdr->cbReturned = sIrp.cbOutput;
    //OSA_DPRINTF(("OsaLinux: OsaLinuxIoDeviceControlEntry returning %d bytes\n", pHdr->cbReturned));
    return 0;
  default:
    OSA_EPRINTF(("OsaLinux: OsaLinuxIoDeviceControlEntry returning error (Info = 0x%X)\n", sIrp.cbOutput));
    pHdr->cbReturned = 0;
    return -1;
  }
}


/******************************************************************************
  Entry function exported to the pure "C" OsaDriverEntry.c
******************************************************************************/
unsigned int OsaLinuxPollDeviceEntry(struct file *pFile, void *pContext, struct poll_table_struct *pPollTable)
{
  OsaLinuxBaseDevice *pBaseDevice = reinterpret_cast<OsaLinuxBaseDevice *>(pContext);
  if (pBaseDevice && pBaseDevice->m_eLinuxDeviceType == OSA_LINUX_DEVICE_USER_EVENT) {
    OsaLinuxUserEventDevice *pDevice = reinterpret_cast<OsaLinuxUserEventDevice *>(pBaseDevice);
    pDevice->PollWait(pFile, pPollTable);
    if (pDevice->IsSet()) {
      return POLLIN | POLLRDNORM | POLLOUT | POLLWRNORM;
    }
  }
  return 0;
}


/******************************************************************************
  Entry function exported to the pure "C" OsaDriverEntry.c
******************************************************************************/
void OsaLinuxRemoveDeviceEntry(void *pContext)
{

  OSA_ASSERT(pContext);
  if (pContext) {

    OsaLinuxBaseDevice *pBaseDevice = reinterpret_cast<OsaLinuxBaseDevice *>(pContext);
    switch (pBaseDevice->m_eLinuxDeviceType) {
    case OSA_LINUX_DEVICE_OS_EXT:
    case OSA_LINUX_DEVICE_USER_EVENT:
      {
        OsaLinuxRemoveProcDevice(pBaseDevice->m_pProcEntry);
        delete pBaseDevice;
      }
      break;
    case OSA_LINUX_DEVICE_EXTERNAL:
      {
        OsaLinuxDevice *pExternalDevice = reinterpret_cast<OsaLinuxDevice *>(pContext);
        OsaLinuxRemoveDevice(pExternalDevice);
      }
      break;
    default:
      OSA_DPRINTF(("OsaLinux: Unknown OsaLinux device type encountered, cannot remove\n"));

      break;
    }
  }
}

/******************************************************************************
******************************************************************************/
void OsaDisconnectNotificationCallbacks(
  IN OsaExternalDeviceRep externalDeviceRep
  )
{
  OsaLinuxDevice *pExternalDevice = reinterpret_cast<OsaLinuxDevice *>(externalDeviceRep);
  if (pExternalDevice->m_pResources) {

    // Disconnect the interrupt first if connected
    if (pExternalDevice->m_bInterruptConnected) {
      osalinux_free_irq(pExternalDevice->m_pResources->sInterrupt.nLevel, pExternalDevice);
      pExternalDevice->m_bInterruptConnected = false;
      OSA_DPRINTF(("OsaLinux: Interrupt %d Disconnected\n", pExternalDevice->m_pResources->sInterrupt.nLevel));
    } else {
      OSA_DPRINTF(("OsaLinux: No interrupt to disconnect\n"));
    }
  }
}

/******************************************************************************
******************************************************************************/
bool OsaConnectNotificationCallbacks(
  IN OsaExternalDeviceRep              externalDeviceRep,
  IN OsaBoardNotificationFnP           pfnBoardNotificationCallback,
  IN OsaDeferredBoardNotificationFnP   pfnDeferredBoardNotificationCallback,
  IN OsaResourcesS::InterruptResourceS *pInterruptResource
  )
{
  OsaLinuxDevice *pExternalDevice = reinterpret_cast<OsaLinuxDevice *>(externalDeviceRep);

  pExternalDevice->m_pfnDeferredBoardNotificationCallback = pfnDeferredBoardNotificationCallback;
  pExternalDevice->m_pfnBoardNotificationCallback = pfnBoardNotificationCallback;

  if (g_pfnOsaLinuxCsIoControl) {
    // PCMCIA support
    OSA_DPRINTF(("OsaLinux: About to call OTX_DRV_CS_IO_SET_INTERRUPT_CALLBACK in PCMCIA CS client driver\n"));

    int nCsIoCtlRc;
    OsaLinuxCsIoInSetInterruptCallbackS sIn;

    sIn.pfnBoardNotificationCallback = OsaLinuxInterruptCallback;
    nCsIoCtlRc = (*g_pfnOsaLinuxCsIoControl)(pExternalDevice->m_pDeviceCompanionDriver, OTX_DRV_CS_IO_SET_INTERRUPT_CALLBACK, &sIn, sizeof(sIn), NULL, 0);
    if (nCsIoCtlRc) {
      OSA_EPRINTF(("OsaLinux: OTX_DRV_CS_IO_SET_INTERRUPT_CALLBACK call failed. nCsIoCtlRc = %d \n", nCsIoCtlRc));
      return false;
    }
    OSA_DPRINTF(("OsaLinux: OTX_DRV_CS_IO_SET_INTERRUPT_CALLBACK call was successful\n"));

  } else {
    // PCI
    OSA_DPRINTF(("OsaLinux: Driver was started without the 'nfnCsIoControl' parameter. Skipping call to PCMCIA CS client driver.\n"));

    if (!pExternalDevice->m_pResources->sInterrupt.bHaveInterrupt) {
      OSA_DPRINTF(("OsaLinux: !pExternalDevice->m_pResources->sInterrupt.bHaveInterrupt!\n"));
      return false;
    }

    int nError = osalinux_request_irq(pExternalDevice->m_pResources->sInterrupt.nLevel, 
                                      pExternalDevice);

    if (!nError) {
      OSA_DPRINTF(("OsaLinux: Interrupt %d connected\n", pExternalDevice->m_pResources->sInterrupt.nLevel));
      pExternalDevice->m_bInterruptConnected = true;
    } else {
      OSA_EPRINTF(("OsaLinux: Unexpected request_irq() error\n"));
      return false;
    }

  }

  return true;
}


/******************************************************************************
******************************************************************************/
/*
int OsaLinuxStartFirstPcmciaBoard()
{
  int status = -1;
  int nCsIoCtlRc;

  OSA_DPRINTF(("OsaLinux: Starting one OTX PCMCIA card\n"));

#if 0 // Temporary
  OSA_DPRINTF(("OsaLinux: Temporary returning here\n"));
  return status;
#endif

  // Defensive check
  if (!g_pfnOsaLinuxCsIoControl) {
    OSA_DPRINTF(("OsaLinux: Driver was started without the 'nfnCsIoControl' parameter. Skipping PCMCIA enumeration.\n"));
    return status;
  }

  OsaLinuxCsIoOutGetVersionS sCsVersion;
  nCsIoCtlRc = (*g_pfnOsaLinuxCsIoControl)(NULL, OTX_DRV_CS_IO_GET_VERSION, NULL, 0, &sCsVersion, sizeof(sCsVersion));
  if (nCsIoCtlRc) {
    OSA_EPRINTF(("OsaLinux: OTX_DRV_CS_IO_GET_VERSION call failed. nCsIoCtlRc = %d \n", nCsIoCtlRc));
    return status;
  }

  // Check that the interface versions match (between the otx_cs and the OtxDrv drivers).
  if (sCsVersion.nVersion != OTX_DRV_CS_VERSION) {
    OSA_DPRINTF(("OsaLinux: Version mismatch in interface to CS driver. CS driver has version %d. We have version %d.\n", sCsVersion.nVersion, OTX_DRV_CS_VERSION));
    return status;
  }

  // Notify the PCMCIA client driver about our OsaLinuxAddDeviceCallback() function.
  // The PCMCIA client driver will call this function for any added OTX PCMCIA cards
  // (but not for the first card (see comment below)).
  OsaLinuxCsIoInSetCallbacksS sCallbacks;
  sCallbacks.pfnAddDeviceCallback = OsaLinuxAddDeviceCallback;
  sCallbacks.pfnRemoveDeviceCallback = OsaLinuxRemoveDeviceCallback;
  nCsIoCtlRc = (*g_pfnOsaLinuxCsIoControl)(NULL, OTX_DRV_CS_IO_SET_CALLBACKS, &sCallbacks, sizeof(sCallbacks), NULL, 0);
  if (nCsIoCtlRc) {
    OSA_EPRINTF(("OsaLinux: OTX_DRV_CS_IO_SET_CALLBACKS call failed. nCsIoCtlRc = %d \n", nCsIoCtlRc));
    return false;
  }
  OSA_DPRINTF(("OsaLinux: OTX_DRV_CS_IO_SET_CALLBACKS call was successful\n"));


  //
  // Since the OTX Driver is not loaded when the PCMCIA driver is processing the CARD INSERTED
  // event for the first card, it can not call the OsaLinuxAddDeviceCallback() function at that
  // time. Therefore we have to "simulate" the AddDevice call when the driver loads.
  //

  OSA_DPRINTF(("OsaLinux: Getting the device pointer for the card device from the PCMCIA client driver\n"));
  OsaLinuxCsIoOutGetCsDeviceS sCsDevice;
  nCsIoCtlRc = (*g_pfnOsaLinuxCsIoControl)(NULL, OTX_DRV_CS_IO_GET_CS_DEVICE, NULL, NULL, &sCsDevice, sizeof(sCsDevice));
  if (nCsIoCtlRc) {
    OSA_EPRINTF(("OsaLinux: OTX_DRV_CS_IO_GET_CS_DEVICE call failed. nCsIoCtlRc = %d \n", nCsIoCtlRc));
    return status;
  }
  OSA_DPRINTF(("OsaLinux: OTX_DRV_CS_IO_GET_CS_DEVICE call was successful\n"));


  return OsaLinuxAddDeviceCallback(sCsDevice.pCsDevice);

}
*/

/******************************************************************************

  Plug-n-Play function.

  Called (by the PCMCIA client driver) when a new board is inserted.

******************************************************************************/
int OsaLinuxAddDeviceCallback(
  IN OsaLinuxCsIoControlFnP pfnIoControl,
  IN void *pCompanionDevice
)
{
  int status = -1;
  int nCsIoCtlRc;
  OsaUint16 nVid; // 0x14C9
  OsaUint16 nDid; // 1012 for Thor-2-PCMCIA;
  OsaUint16 nSid; // 0x0420 for 1.0;

  OSA_DPRINTF(("OsaLinux: OsaLinuxAddDeviceCallback() called. pCompanionDevice=0x%p\n", pCompanionDevice));

  if (!pfnIoControl) {
    OSA_DPRINTF(("OsaLinux: OsaLinuxAddDeviceCallback() pfnIoControl is NULL. Ignoring this call.\n"));
    return status;
  }
  g_pfnOsaLinuxCsIoControl = pfnIoControl;

  OsaLinuxCsIoOutGetVersionS sCsVersion;
  nCsIoCtlRc = (*pfnIoControl)(NULL, OTX_DRV_CS_IO_GET_VERSION, NULL, 0, &sCsVersion, sizeof(sCsVersion));
  if (nCsIoCtlRc) {
    OSA_EPRINTF(("OsaLinux: OTX_DRV_CS_IO_GET_VERSION call failed. nCsIoCtlRc = %d \n", nCsIoCtlRc));
    return status;
  }

  // Check that the interface versions match (between the otx_cs and the OtxDrv drivers).
  if (sCsVersion.nVersion != OTX_DRV_CS_VERSION) {
    OSA_DPRINTF(("OsaLinux: Version mismatch in interface to CS driver. CS driver has version %d. We have version %d.\n", sCsVersion.nVersion, OTX_DRV_CS_VERSION));
    return status;
  }


  // Get the product number of the PCMCIA card (dev_name)
  OsaLinuxCsIoOutGetDevNameS sCsDevName;
  nCsIoCtlRc = (*g_pfnOsaLinuxCsIoControl)(pCompanionDevice, OTX_DRV_CS_IO_GET_DEV_NAME, NULL, 0, &sCsDevName, sizeof(sCsDevName));
  if (nCsIoCtlRc) {
    OSA_EPRINTF(("OsaLinux: OTX_DRV_CS_IO_GET_DEV_NAME call failed. nCsIoCtlRc = %d \n", nCsIoCtlRc));
    return status;
  }
  OSA_DPRINTF(("OsaLinux: CS driver reports a device with device name %s.\n", sCsDevName.szDevName));


  // Check which type of OTX PCMCIA card it is
  if (strstr(sCsDevName.szDevName, "HAA-1012-1-1.0")) {
    nVid = OTX_ODIN_VENDOR_ID;
    nDid = OTX_PROD_ID_THOR_2_PCMCIA_STD;
    nSid = 0x0420;  // @todo should be macro OD_NIC_SUBSYSTEM_ID_REV_1_0

  } else if (strstr(sCsDevName.szDevName, "HAA-1012-1-1.1")) {
    nVid = OTX_ODIN_VENDOR_ID;
    nDid = OTX_PROD_ID_THOR_2_PCMCIA_STD;
    nSid = 0x0460;  // @todo should be macro OD_NIC_SUBSYSTEM_ID_REV_1_1

  } else if (strstr(sCsDevName.szDevName, "HAA-1043-1-1.0-P1")) {
    nVid = OTX_ODIN_VENDOR_ID;
    nDid = OTX_PROD_ID_THOR_2_PCMCIA_PLUS;
    nSid = 0x0401;  // @todo should be macro OD_NIC_SUBSYSTEM_ID_REV_1_0_P1

  } else if (strstr(sCsDevName.szDevName, "HAA-1043-1-1.0-P2")) {
    nVid = OTX_ODIN_VENDOR_ID;
    nDid = OTX_PROD_ID_THOR_2_PCMCIA_PLUS;
    nSid = 0x0402;  // @todo should be macro OD_NIC_SUBSYSTEM_ID_REV_1_0_P2

  } else if (strstr(sCsDevName.szDevName, "HAA-1043-1-1.0")) {
    nVid = OTX_ODIN_VENDOR_ID;
    nDid = OTX_PROD_ID_THOR_2_PCMCIA_PLUS;
    nSid = 0x0420;  // @todo should be macro OD_NIC_SUBSYSTEM_ID_REV_1_0

  } else if (strstr(sCsDevName.szDevName, "HAA-1012-1")) {
    // Original Thor-2-PCMCIA 1.0
    nVid = OTX_ODIN_VENDOR_ID;
    nDid = OTX_PROD_ID_THOR_2_PCMCIA_STD;
    nSid = 0x0420;  // @todo should be macro OD_NIC_SUBSYSTEM_ID_REV_1_0

  } else if (strstr(sCsDevName.szDevName, "HAA-1057-1-1.0")) {
    nVid = OTX_ODIN_VENDOR_ID;
    nDid = OTX_PROD_ID_THOR_2_PCMCIA_CST;
    nSid = 0x0420;  // @todo should be macro OD_NIC_SUBSYSTEM_ID_REV_1_0

  } else if (strstr(sCsDevName.szDevName, "HAA-1058-1-1.0")) {
    nVid = OTX_ODIN_VENDOR_ID;
    nDid = OTX_PROD_ID_THOR_2_PCMCIA_EX;
    nSid = 0x0420;  // @todo should be macro OD_NIC_SUBSYSTEM_ID_REV_1_0

  } else if (strstr(sCsDevName.szDevName, "HAA-1060-1-1.0")) {
    nVid = OTX_ODIN_VENDOR_ID;
    nDid = OTX_PROD_ID_THOR_2_PCMCIA_PRO;
    nSid = 0x0420;  // @todo should be macro OD_NIC_SUBSYSTEM_ID_REV_1_0

  } else {
    // Unknown Odin PCMCIA board
    OSA_DPRINTF(("OsaLinux: Unknown Odin PCMCIA board (%s)\n", sCsDevName.szDevName));
    return status;
  }


  OsaLinuxDevice *pExternalDevice = NULL;

  if (!OsaLinuxAddDevice(
    OSA_BUS_PCMCIA,
    nVid,
    nDid,
    nSid,
    NULL,
    OUT &pExternalDevice
    ))
  {
    // Probably not a device the Driver wanted to handle
    OSA_DPRINTF(("OsaLinux: Probably not a device the Driver wanted to handle\n"));
    return status;
  }

  OSA_DPRINTF(("OsaLinux: Added PCMCIA device (nVid=%X, nDid=%X, nSid=%X)\n", nVid, nDid, nSid));

  pExternalDevice->m_pResources = new OsaResourcesS;

  if (!pExternalDevice->m_pResources) {
    OSA_DPRINTF(("OsaLinux: Out of memory\n"));
    return status;
  }

  OSA_DPRINTF(("OsaLinux: Getting Card IO Base address the PCMCIA client driver\n"));
  OsaLinuxCsIoOutGetIoBaseS sCsIoBase;
  nCsIoCtlRc = (*g_pfnOsaLinuxCsIoControl)(pCompanionDevice, OTX_DRV_CS_IO_GET_IO_BASE, NULL, NULL, &sCsIoBase, sizeof(sCsIoBase));
  if (nCsIoCtlRc) {
    OSA_EPRINTF(("OsaLinux: OTX_DRV_CS_IO_GET_IO_BASE call failed. nCsIoCtlRc = %d \n", nCsIoCtlRc));
    return status;
  }
  OSA_DPRINTF(("OsaLinux: OTX_DRV_CS_IO_GET_IO_BASE call was successful\n"));

  if (sCsIoBase.nSlot < 0) {
    // Unexpected. We don't have any cards in the PCMCIA socket
    OSA_EPRINTF(("OsaLinux: Unexpected. We dont have any cards in the PCMCIA socket\n"));
    return status;
  }


  pExternalDevice->m_pResources->coPortAreas = 2;
  pExternalDevice->m_pResources->coMemoryAreas = 0;
  pExternalDevice->m_pResources->sqIoPortArea[0].cbLength          = sCsIoBase.nSize1;
  pExternalDevice->m_pResources->sqIoPortArea[0].pStart.sWord.nLow = sCsIoBase.nIobase1;
  pExternalDevice->m_pResources->sqIoPortArea[1].cbLength          = sCsIoBase.nSize2;
  pExternalDevice->m_pResources->sqIoPortArea[1].pStart.sWord.nLow = sCsIoBase.nIobase2;
  pExternalDevice->m_bInterruptConnected = false;  // The PCMCIA Client driver handles the interrupt
                                               // connection/disconnection. Therefore we set
                                               // this variable to false.
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
  pExternalDevice->m_bSoftIsrActive = false; // used to show that the soft ISR is active
#endif

#if 1 // Needed since we perform a check on this variable in OdPnpDevice
  pExternalDevice->m_pResources->sInterrupt.bHaveInterrupt = true;  // Setting this variable to true will tell the PnPDevice to
                                                                    // connect the Notification Callback routine
#endif
  OSA_ASSERT(pExternalDevice->m_pResources->sqIoPortArea[0].cbLength == 32);
  OSA_ASSERT(pExternalDevice->m_pResources->sqIoPortArea[1].cbLength == 16);



  OSA_DPRINTF(("OsaLinux: Provide our pExternalDevice to the PCMCIA client driver\n"));

  OsaLinuxCsIoInSetExternalDeviceS sIn;

  sIn.pExternalDevice = (void *)pExternalDevice;
  nCsIoCtlRc = (*g_pfnOsaLinuxCsIoControl)(pCompanionDevice, OTX_DRV_CS_IO_SET_EXTERNAL_DEVICE, &sIn, sizeof(sIn), NULL, 0);
  if (nCsIoCtlRc) {
    OSA_EPRINTF(("OsaLinux: OTX_DRV_CS_IO_SET_EXTERNAL_DEVICE call failed. nCsIoCtlRc = %d \n", nCsIoCtlRc));
    return status;
  }

  OSA_DPRINTF(("OsaLinux: OTX_DRV_CS_IO_SET_EXTERNAL_DEVICE call was successful\n"));



  // Start the board using the resources supplied by the PCMCIA CS driver
  // @todo Make sure that the OTX driver does not try to free these resources (e.g. io regions and irqs)
  //       since this is a task for the PCMCIA CS driver
  if (!OSA_SUCCESS((*g_sOsaDriverEntryPoints.pfnStartBoard)(pExternalDevice->m_internalDeviceRep, pExternalDevice, pExternalDevice->m_pResources))) {
    // @todo Maybe we need to remove the device if this call was unsuccessful (similar to OsaNtRemoveDevice())
    return status;
  }

  status = 0;

  // Keep track of how many devices that is using the OTX driver (to prevent rmmod from
  // removing it when it is still being used)
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
  OsaLinux_MOD_INC_USE_COUNT();
#endif

  return status;


}


/******************************************************************************

  Plug-n-Play function

  Called (by the PCMCIA Client driver) when a board is removed.

  Exported to pure "C".

******************************************************************************/
int OsaLinuxRemoveDeviceCallback(
  IN void *pExternalDevice
)
{
  bool rc;

  rc = OsaLinuxRemoveDevice( (OsaLinuxDevice *)pExternalDevice);

  // Keep track of how many devices that is using the OTX driver (to prevent rmmod from
  // removing it when it is still being used)
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
  if (rc) {
    OsaLinux_MOD_DEC_USE_COUNT();
  }
#endif
 /* FIXME: we should do this in other way if we support > 1 pcmcia boards */
 g_pfnOsaLinuxCsIoControl = NULL;

  return (int) rc;
}




/******************************************************************************
******************************************************************************/
int OsaLinuxEnumAndAddAndStartAllPciBoards()
{
  int status = 0;  // '0' = Success

  OSA_DPRINTF(("OsaLinux: Enumerating PCI devices\n"));

#ifdef CONFIG_PCI
  // Note: All unix like identifiers (non-mixed case and underscore) are from the Linux kernel headers
  if (OsaLinux_pci_present()) {

    void *pPciDev;
    for (pPciDev = OsaLinux_pci_dev_next(NULL); pPciDev != NULL; pPciDev = OsaLinux_pci_dev_next(pPciDev)) {

      if(pPciDev == NULL) {
        OSA_DPRINTF(("OsaLinux: pPciDev NULL - Search Finished\n\n"));

        return(status);
      }
      // At this point we have found a board

      // Declare a functional device object (as opposed to a physical device object which is
      // present lower in the bus driver in WDM.
      // The object is allocated by Osa on request by Od
      OsaLinuxDevice *pExternalDevice = NULL;
      OsaInternalDeviceRep internalDeviceRep = NULL;

      OsaUint16 SubSystemID;
      OsaLinux_pci_read_config_word(pPciDev, PCI_SUBSYSTEM_ID, &SubSystemID);

      if (!OsaLinuxAddDevice(
        OSA_BUS_PCI,
        OsaLinux_pci_dev_vendor(pPciDev),
        OsaLinux_pci_dev_device(pPciDev),
        SubSystemID,
        NULL,
        OUT &pExternalDevice
        ))
      {
        // Probably not a device the Driver wanted to handle
       OSA_DPRINTF(("OsaLinux: Not a device the driver supports\n"));

       continue;
      }

      OSA_DPRINTF(("OsaLinux: Found and added PCI device\n"));

      OsaResourcesS *pOsaResources = NULL;

      pExternalDevice->m_pPciDev = pPciDev;

      OsaLinux_pci_dev_enable(pPciDev);
      // Claim resources, OsaLinuxRemoveDevice frees the resources (by calling OsaLinuxReturnResources)
      if (!OsaLinuxExtractPciResources(
        pPciDev,
        OUT &pOsaResources      // If the board is requesting resources we get the requests here (allocated with new)
        ))
      {
        // If error code we assume the AssignResources did not allocate pOsaResources for us
        // If error we should not get any resources allocated for us so we don't need to delete them
        OSA_EPRINTF(("OsaLinux: Failed to extract resources\n"));
        OSA_ASSERT(!pOsaResources);
        OsaLinuxRemoveDevice(pExternalDevice);
        continue;
      }

      OSA_DPRINTF(("OsaLinux: Extracted resources\n"));

      OSA_ASSERT(pOsaResources);  // All our boards request resources (how would the be used otherwise?)

      bool bResult = OsaLinuxStartDevice(pExternalDevice, pOsaResources);

      delete pOsaResources; // These are only the unmapped and unclaimed resources

      if (!bResult) {
        // If the device could not be started we must delete it
        // Any resources claimed will be released by OsaLinuxRemoveDevice calling OsaLinuxReturnResources
        OSA_DPRINTF(("<1>Failed to start OTX PCI device\n"));
        OsaLinuxRemoveDevice(pExternalDevice);
        status = -1;
        continue;
      } else {
        OSA_DPRINTF(("OsaLinux: Started device\n"));
      }

    }

  } else {
    OSA_DPRINTF(("<1>No PCI support in kernel, loading aborted\n"));
    return -1;
  }

  return status;
#else
  OSA_DPRINTF(("<1>No PCI support in kernel, loading aborted\n"));
  return -1;
#endif
}


/******************************************************************************
******************************************************************************/
bool OsaLinuxAddDevice(
  IN OsaBusTypeE  eBusType,
  IN OsaUint16 nVid,
  IN OsaUint16 nDid,
  IN OsaUint16 nSid,
  IN wchar_t *pwcszVendorString,
  OUT OsaLinuxDevice **ppExternalDevice
)
{
  *ppExternalDevice = NULL;
  OsaString strExternalDeviceName;
  OsaInternalDeviceRep internalDeviceRep = NULL;
  OsaLinuxDevice *pExternalDevice = NULL;

  // We now support both PCI and PCMCIA
  if (!(eBusType == OSA_BUS_PCI || eBusType == OSA_BUS_PCMCIA)) {
    OSA_EPRINTF(("OsaLinux: Unsupported Bus Type %d\n", eBusType));
    return false;
  }

  if((nDid & (OsaUint16) 0xF000) == (OsaUint16) 0xF000) {
    // Thor-2-ExpressCard Bridge use F432 as its device ID. This
    // is not a valid device for the OTX driver. Use the 0xF000  
    // mask to determine a valid device ID. No Odin Device ID  
    // contains an 'F' in the MSB. Found in driver release 2.18.1.0.
    OSA_DPRINTF(("OsaLinux: Unsupported Device ID %d(0x%X) on Bus Type %d \n",
                 nDid, 
                 nDid, 
                 eBusType));
    return false;
  }

  if((nSid & (OsaUint16) 0xFFFF) == (OsaUint16) 0xFFFF) {
    // Thor-2-ExpressCard, if card is installed after the laptop
    // PC has been booted the Sub System ID is set to 0xFFFF. If
    // this is the case then starting the driver would cause the 
    // PC to crash because the card is not initialized correctly.
    // Found in driver release 2.18.1.0.
    OSA_EPRINTF(("OsaLinux: Unsupported SubSystem ID %d(0x%X) "
                 "Device ID 0x%X Bus Type %d \n",
                 nSid, 
                 nSid, 
                 nDid,
                 eBusType));
    return false;
  }

  if (!OSA_SUCCESS((*g_sOsaDriverEntryPoints.pfnAddBoard)(eBusType, nVid, nDid, nSid, pwcszVendorString, &internalDeviceRep, &strExternalDeviceName))) {
    return false;
  }

  if (!OSA_SUCCESS(OsaCreateExternalDevice(
    strExternalDeviceName,                // Externally known name of the board device
    internalDeviceRep,                    // A reference by which the device is known internally (Driver)
    (void **) &pExternalDevice            // A reference by which the device is known externally (OSA and System)
    )))
  {
    // Cleanup after (*AddBoardFnP)() since if OsaLinuxAddDevice fails
    // the caller won't call OsaLinuxRemoveDevice (caller thinks device was never added)
    OSA_EPRINTF(("OsaLinux: Failed to create external device\n", eBusType));
    (*g_sOsaDriverEntryPoints.pfnRemoveBoard)(internalDeviceRep);
    return false;
  }

  OSA_ASSERT(pExternalDevice);
  pExternalDevice->m_bBoard = true;

  *ppExternalDevice = pExternalDevice;

  return true;
}


/******************************************************************************

  Figures out the resources needed for the device

  Cleanup needed: Caller is responsible for deleting *ppResources if return
  value is true

******************************************************************************/
bool OsaLinuxExtractPciResources(
  IN void *pPciDev,
  OUT OsaResourcesS **ppResources
)
{
  OSA_ASSERT(ppResources && !*ppResources);
  OSA_ASSERT(pPciDev);

  // If we return failure we must delete this resources object
  // If we return true we must attach it to the OUT parameter
  OsaResourcesS *pResources = new OsaResourcesS;

  OSA_ASSERT(pResources);

  if (!pResources) {
    return false;
  }


  // Determine what resources to use

  if (OsaLinux_pci_dev_irq(pPciDev)) {
    pResources->sInterrupt.bHaveInterrupt = true;
    pResources->sInterrupt.nLevel = OsaLinux_pci_dev_irq(pPciDev);
    // The interrupt is connected in ConnectNotificationCallbacks()
  }

  pResources->coPortAreas = 0;
  pResources->coMemoryAreas = 0;

  for (int idx = 0; idx <= 5; idx++) {
    u32 nPciBaseAddrRegOffset = PCI_BASE_ADDRESS_0 + 4*idx;

    // Figure out how much space and of what type this
    // device wants.
    //
    if (!OsaLinux_pci_dev_resource_start(pPciDev, idx)) {
      // this base-address register is unused
      // We perhaps should break the loop here since there should not be any more regions according to PCI specs
      continue;
    }

    u32 nPciOrigAddrVal;
    u32 nPciAddr;
    u32 mask;

    // During the following temporary remap of the PCI registers and ports we
    // should prevent any other kernel component from touching PCI memory
    // in a UP architecture this is easy, just disabling the interrups
    // and this 'thread' of execution owns the CPU. On an SMP archtecture
    // I don't know how to make this 'thread' king, now when the kernel
    // is using finer grained spinlocks instead of a big one.
    // The bios32.c kernel file uses a pci_lock, but that only seems
    // to protect config space.
    // I do an global_cli to prevent interrupts on all processors (I think)
    // and then I cross my fingers that no other processor is running an
    // dispatch thread doing PCI access.
    //

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	cli();
#else
    unsigned f;
    osalinux_save_flags_cli(&f);
#endif
    OsaLinux_pci_read_config_dword(pPciDev, nPciBaseAddrRegOffset, &nPciOrigAddrVal);
    OsaLinux_pci_write_config_dword(pPciDev, nPciBaseAddrRegOffset, 0xffffffff);
    OsaLinux_pci_read_config_dword(pPciDev, nPciBaseAddrRegOffset, &nPciAddr);

    // Write back the original value to the PCI board
    OsaLinux_pci_write_config_dword(pPciDev, nPciBaseAddrRegOffset, nPciOrigAddrVal);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	sti();
#else
    osalinux_restore_flags(f);
#endif

    // We've read the base address register back after
    // writing all ones and so now we must decode it.
    //
    if (nPciAddr & PCI_BASE_ADDRESS_SPACE_IO) {

      // I/O space base address register.
      //
      nPciAddr &= PCI_BASE_ADDRESS_IO_MASK;
      mask = (~nPciAddr << 1) | 0x1;

      pResources->sqIoPortArea[pResources->coPortAreas].cbLength = (mask & nPciAddr) & 0xffffffff;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
      // Use the value the kernel has decided for us, not from the PCI device
      pResources->sqIoPortArea[pResources->coPortAreas].pStart =  OsaLinux_pci_dev_resource_start(pPciDev, idx) & PCI_BASE_ADDRESS_IO_MASK;
#else
      pResources->sqIoPortArea[pResources->coPortAreas].pStart.sWord.nLow = OsaLinux_pci_dev_resource_start(pPciDev, idx);
#endif

      pResources->coPortAreas++;
    } else {

      // Memory space base address register.
      //
      nPciAddr &= PCI_BASE_ADDRESS_MEM_MASK;
      mask = (~nPciAddr << 1) | 0x1;

      pResources->sqMemoryRegArea[pResources->coMemoryAreas].cbLength  = (mask & nPciAddr) & 0xffffffff;

      pResources->sqMemoryRegArea[pResources->coMemoryAreas].pStart.sWord.nLow = OsaLinux_pci_dev_resource_start(pPciDev, idx) & PCI_BASE_ADDRESS_MEM_MASK;  //yh: it was  PCI_BASE_ADDRESS_IO_MASK

      pResources->coMemoryAreas++;
    }

  }

  *ppResources = pResources;

  return true;
}


/******************************************************************************
  Starts a functional device object that represents a device on a bus (a board).

  This function is deliberately made similar to IRP_MN_START_DEVICE. However
  the resource list is passed as a distinct parameter an not in the stack of
  an IRP.

  If this function fails the caller must call OsaLinuxRemoveDevice to:
  Cleanup after (*AddBoardFnP)()
  Cleanup after partially mapped and claimed resources
******************************************************************************/
bool OsaLinuxStartDevice(
  IN OsaLinuxDevice *pExternalDevice,
  IN OsaResourcesS *pResources
)
{
  // Create a resource list for the claimed and mapped resources.
  OsaResourcesS *pClaimedResources = new OsaResourcesS;

  OSA_ASSERT(pClaimedResources);

  if (!pClaimedResources) {
    return false;
  }

  // We must attach the resources to the device so they get released with the device
  // We will soon start claiming resources and may return with an error
  // Any error returned by this function will result in a call to OsaLinuxRemoveDevice and OsaLinuxReturnResources
  // Return resources will return any resources we have claimed and mapped here, but only
  // if they can be found through pExternalDevice->m_pResources
  pExternalDevice->m_pResources = pClaimedResources;  // We have no resources claimed yet, but an empty OsaResourcesS object is okay too

  pClaimedResources->sInterrupt.bHaveInterrupt = pResources->sInterrupt.bHaveInterrupt;
  pClaimedResources->sInterrupt.nLevel = pResources->sInterrupt.nLevel;
  pClaimedResources->coPortAreas = 0;
  // Claim the I/O ports
  for (int iPort = 0; iPort < pResources->coPortAreas; iPort++) {
    OSA_DPRINTF(("OsaLinux: PCI I/O area length = %d\n", pResources->sqIoPortArea[iPort].cbLength));
    OSA_DPRINTF(("OsaLinux: PCI I/O area start (raw)= 0x%x\n", (OsaUint32) pResources->sqIoPortArea[iPort].pStart));

    if (check_region(pResources->sqIoPortArea[iPort].pStart, pResources->sqIoPortArea[iPort].cbLength)) {
      // Region taken, since it's PCI which was setup by the kernel/bios this should not happen
      OSA_EPRINTF(("OsaLinux: Unexpected check_region() error\n"));
      return false;
    }

    // Always succeeeds if the check_region() above succeeded
    request_region(pResources->sqIoPortArea[iPort].pStart, pResources->sqIoPortArea[iPort].cbLength, "OSA_Device");

    pClaimedResources->sqIoPortArea[iPort].pStart = pResources->sqIoPortArea[iPort].pStart;
    pClaimedResources->sqIoPortArea[iPort].cbLength = pResources->sqIoPortArea[iPort].cbLength;
    pClaimedResources->coPortAreas = iPort + 1;

    OSA_DPRINTF(("OsaLinux: PCI I/O area start (claimed)= 0x%x\n", (OsaUint32) pClaimedResources->sqIoPortArea[iPort].pStart));

  }

  // Map the physical memory addresses to virtual addresses
  pClaimedResources->coMemoryAreas = 0;
  for (int iMemory = 0; iMemory < pResources->coMemoryAreas; iMemory++) {

    OSA_DPRINTF(("OsaLinux: PCI Memory area length = %d\n", pResources->sqMemoryRegArea[iMemory].cbLength));
    OSA_DPRINTF(("OsaLinux: PCI Memory area start (unmapped)= 0x%x\n", (OsaUint32) pResources->sqMemoryRegArea[iMemory].pStart));


    pClaimedResources->sqMemoryRegArea[iMemory].pStart = osalinux_ioremap(pResources->sqMemoryRegArea[iMemory].pStart, pResources->sqMemoryRegArea[iMemory].cbLength);
    pClaimedResources->sqMemoryRegArea[iMemory].cbLength = pResources->sqMemoryRegArea[iMemory].cbLength;
    pClaimedResources->coMemoryAreas = iMemory + 1;
    if (pClaimedResources->sqMemoryRegArea[iMemory].pStart) {
      pClaimedResources->sqMemoryRegArea[iMemory].eAddressSpace = OsaResourcesS::MemorySpaceMapped;
      OSA_DPRINTF(("OsaLinux: PCI Memory area start (mapped)= 0x%x\n", (OsaUint32) pClaimedResources->sqMemoryRegArea[iMemory].pStart));
    } else {
      // Mapping failed
      pClaimedResources->sqMemoryRegArea[iMemory].cbLength = 0;
      OSA_EPRINTF(("OsaLinux: Unexpected map error\n"));
      return false;
    }
  }

  OSA_DPRINTF(("OsaLinux: %d I/O areas requested\n", pClaimedResources->coPortAreas));
  OSA_DPRINTF(("OsaLinux: %d Mem areas mapped\n", pClaimedResources->coMemoryAreas));

  // Start the board using the claimed and mapped resources.
  if (!OSA_SUCCESS((*g_sOsaDriverEntryPoints.pfnStartBoard)(pExternalDevice->m_internalDeviceRep, pExternalDevice, pExternalDevice->m_pResources))) {
    return false;
  }


  return true;
}


/******************************************************************************
  Removes a functional device object that represents a device on a bus (a board).

  This function is deliberately made similar to the OsaWdmRemoveDevice().

******************************************************************************/
bool OsaLinuxRemoveDevice(
  IN OsaLinuxDevice *pExternalDevice
)
{
  OSA_DPRINTF(("OsaLinux: Removing device\n"));
  // When removing in a Linux unload routine we have no choice, so the driver must be ready
  OsaResultE result = (*g_sOsaDriverEntryPoints.pfnBoardState)(pExternalDevice->m_internalDeviceRep, OSA_BOARD_CAN_REMOVE);
  OSA_ASSERT(OSA_SUCCESS(result));

  (*g_sOsaDriverEntryPoints.pfnBoardState)(pExternalDevice->m_internalDeviceRep, OSA_BOARD_REJECT_NEW_IRPS);
  (*g_sOsaDriverEntryPoints.pfnBoardState)(pExternalDevice->m_internalDeviceRep, OSA_BOARD_WAIT_FOR_IRPS_COMPLETED);
  (*g_sOsaDriverEntryPoints.pfnBoardState)(pExternalDevice->m_internalDeviceRep, OSA_BOARD_DISABLE_IO);

  (*g_sOsaDriverEntryPoints.pfnRemoveBoard)(pExternalDevice->m_internalDeviceRep);

  OsaLinuxReturnResources(pExternalDevice);

  OsaDestroyExternalDevice(pExternalDevice);

  return true;
}



/******************************************************************************
  This is function is called to return any hardware resources which
  have been allocated for the device.  In our case, these are
  the Interrupt Object, and any ports we might have mapped through
  memory, and all assigned resources.  Note that we do NOT return our
  Device Object here.
******************************************************************************/
void OsaLinuxReturnResources(
  OsaLinuxDevice *pExternalDevice
)
{
  OSA_DPRINTF(("OsaLinux: Returning Resources\n"));
  if (pExternalDevice->m_pResources) {

    if (g_pfnOsaLinuxCsIoControl) {
      // PCMCIA support
      OSA_DPRINTF(("OsaLinux: Calling OTX_DRV_CS_IO_CLR_CALLBACKS in PCMCIA CS client driver\n"));

      int nCsIoCtlRc;

      nCsIoCtlRc = (*g_pfnOsaLinuxCsIoControl)(pExternalDevice->m_pDeviceCompanionDriver, OTX_DRV_CS_IO_CLR_CALLBACKS, NULL, 0, NULL, 0);
      OSA_DPRINTF(("OsaLinux: OTX_DRV_CS_IO_CLR_CALLBACKS call returned nCsIoCtlRc = %d \n", nCsIoCtlRc));

      OSA_DPRINTF(("OsaLinux: The IO resources will be released by the PCMCIA card driver. Skipping release/unmap functions here\n"));

    } else {
      // PCI

      OSA_DPRINTF(("OsaLinux: %d I/O areas to release\n", pExternalDevice->m_pResources->coPortAreas));
      OSA_DPRINTF(("OsaLinux: %d Mem areas to unmap\n", pExternalDevice->m_pResources->coMemoryAreas));

      // Now unmap memory address resources and release io ports
      for (int iPort = 0; iPort < pExternalDevice->m_pResources->coPortAreas; iPort++) {
        OSA_DPRINTF(("OsaLinux: Releasing I/O region 0x%p\n", pExternalDevice->m_pResources->sqIoPortArea[iPort].pStart.pVoid));
        release_region(pExternalDevice->m_pResources->sqIoPortArea[iPort].pStart, pExternalDevice->m_pResources->sqIoPortArea[iPort].cbLength);
      }

      for (int iMemory = 0; iMemory < pExternalDevice->m_pResources->coMemoryAreas; iMemory++) {
        OSA_DPRINTF(("OsaLinux: Unmapping memory region 0x%p\n", pExternalDevice->m_pResources->sqMemoryRegArea[iMemory].pStart.pVoid));
        osalinux_iounmap(pExternalDevice->m_pResources->sqMemoryRegArea[iMemory].pStart);
      }

    }

    delete pExternalDevice->m_pResources;
    pExternalDevice->m_pResources = NULL;


  } else {
    OSA_DPRINTF(("OsaLinux: No resources to return\n"));
  }
}


/******************************************************************************
******************************************************************************/
int OsaWaitForSingleEvent(OsaEventImpl *pEvent, OsaUint32 nWaitMaxMs)
{
  int nRemainingMs;
  int rc = 0;

  if (!pEvent->m_nEventSet) {
    // Yes, I have thought of the race condition that occurs after we have
    // tested the the m_nEventSet between a SetEvent call and the osalinux_wait_event_timeout
    // call below, however, osalinux_wait_event_timeout tests the m_nEventSet again in
    // a race safe way so that we don't miss the SetEvent call and sleep forever
    rc =  osalinux_wait_event_timeout(&pEvent->m_nEventSet, &pEvent->m_pWaitQHead, nWaitMaxMs);
  }

  return rc;
}

/******************************************************************************
******************************************************************************/
int OsaWaitForSemaphore(OsaSemaphoreImpl *pSemaphore)
{
  OsaSemaphoreImpl_WaitForSemaphore(&pSemaphore->m_Semaphore);
  return 0;
}


/******************************************************************************
******************************************************************************/
OsaResultE OsaGetKernelRefForEventHandle(OsaVoidP hEvent, OsaUserEventP *ppUserEvent)
{
  //OSA_DPRINTF(("OsaLinux: WARNING: OsaGetKernelRefForEventHandle Not implemented\n"));
  // Although we have a device representing the event object living in OSA the user lib
  // actually keps track of the pointer to this device as an opaque pointer as well
  // as the OS handle to the device. In this case we get the opaque pointer passed
  // to us so we just need to cast it.
  *ppUserEvent = (OsaUserEventP) hEvent;

  OSA_DPRINTF(("OsaLinux: OsaGetKernelRefForEventHandle: Event Pointer = %p\n", *ppUserEvent));

  return OSA_RES_SUCCESS;
}

/******************************************************************************
******************************************************************************/
OsaResultE OsaCloseKernelRefForEventHandle(OsaUserEventP ppUserEvent)
{
  return OSA_RES_SUCCESS;
}

/******************************************************************************
******************************************************************************/
OsaBool32 OsaSetUserEvent(OsaUserEventP pUserEvent)
{
  OsaLinuxUserEventDevice *pEvent = reinterpret_cast<OsaLinuxUserEventDevice *>(pUserEvent);

  // Don't forget the user event must be "auto reset" whereas kernel events
  // are "manual reset"
  // Auto-reset is not possible so we have changed the EventQueue to always
  // reset the pointer when extracting from and empty fifo
  if (pEvent) {
    //OSA_DPRINTF(("OsaLinux: OsaSetUserEvent: Event Pointer = %p\n", pEvent));
    if (pEvent->m_pClassValidityObject != (void *) &OsaLinuxUserEventDevice::ms_classValidityObject) {
      OSA_DPRINTF(("OsaSetUserEvent: Not a valid event pointer\n"));
      return OSA_TRUE;
    } else {
      //OSA_DPRINTF(("OsaSetUserEvent: Valid pointer to an event\n"));
      pEvent->SetEvent();
    }
  }

  return OSA_FALSE;
}


/******************************************************************************
******************************************************************************/
void OsaClearUserEvent(OsaUserEventP pUserEvent)
{
  OsaLinuxUserEventDevice *pEvent = reinterpret_cast<OsaLinuxUserEventDevice *>(pUserEvent);

  // Don't forget the user event must be "auto reset" whereas kernel events
  // are "manual reset"
  if (pEvent) {
    //OSA_DPRINTF(("OsaLinux: OsaClearUserEvent: Event Pointer = %p\n", pEvent));
    if (pEvent->m_pClassValidityObject != (void *) &OsaLinuxUserEventDevice::ms_classValidityObject) {
      OSA_DPRINTF(("OsaClearUserEvent: Not a valid event pointer\n"));
    } else {
      //OSA_DPRINTF(("OsaClearUserEvent: Valid pointer to an event\n"));
      pEvent->ClearEvent();
    }
  }
}


/******************************************************************************
******************************************************************************/
bool OsaDmaAdapterCheck(IN OsaExternalDeviceRep pBoard)
{
  OsaLinuxDevice *pExternalDevice = reinterpret_cast<OsaLinuxDevice *>(pBoard);

  return true;
}


/******************************************************************************
******************************************************************************/
bool OsaDmaInitAdapter(IN OsaExternalDeviceRep pBoard, unsigned long ulMaxLength, unsigned long ulMaxFragments)
{
  OsaLinuxDevice *pExternalDevice = reinterpret_cast<OsaLinuxDevice *>(pBoard);

  pExternalDevice->m_DmaMaxMapRegisters = ulMaxFragments;
  return true;
}


/******************************************************************************
******************************************************************************/
bool OsaDmaUnInitAdapter(IN OsaExternalDeviceRep pBoard)
{
  OsaLinuxDevice *pExternalDevice = reinterpret_cast<OsaLinuxDevice *>(pBoard);

  return true;
}

/******************************************************************************
******************************************************************************/
unsigned long OsaGetDmaMaxMapRegisters(IN OsaExternalDeviceRep pBoard)
{
  OsaLinuxDevice *pExternalDevice = reinterpret_cast<OsaLinuxDevice *>(pBoard);

  return (unsigned long) pExternalDevice->m_DmaMaxMapRegisters;
}

/******************************************************************************
******************************************************************************/
bool OsaDmaAllocateAdapter(
  IN OsaExternalDeviceRep pBoard,
  IN void * Context
)
{
  OsaLinuxDevice *pExternalDevice = reinterpret_cast<OsaLinuxDevice *>(pBoard);

  return true;
}

/******************************************************************************
******************************************************************************/
bool OsaDmaFreeAdapter(
  IN OsaExternalDeviceRep pBoard,
  IN void * Context
)
{
  OsaLinuxDevice *pExternalDevice = reinterpret_cast<OsaLinuxDevice *>(pBoard);

  return true;
}

/******************************************************************************
******************************************************************************/
bool OsaDmaAllocateCommonBuffer(IN OsaExternalDeviceRep pBoard, IN OUT POSA_MD pOmd)
{
  OsaLinuxDevice *pExternalDevice = reinterpret_cast<OsaLinuxDevice *>(pBoard);

  pOmd->pLog = osalinux_alloc_dma_common_buf(pExternalDevice->m_pPciDev, pOmd->cbSize, (unsigned int *) &pOmd->pPhy);
  if (!pOmd->pLog) {
    pOmd->pPhy = 0;
  }

  pOmd->pMdl     = (void *) NULL;  // If not used set to NULL
  pOmd->pContext = (void *) NULL;
  pOmd->ui64Addr = 0;

  return (pOmd->pLog != (void *) NULL);
}

/******************************************************************************
******************************************************************************/
bool OsaDmaFreeCommonBuffer(IN OsaExternalDeviceRep pBoard, IN OUT POSA_MD pOmd)
{
  OsaLinuxDevice *pExternalDevice = reinterpret_cast<OsaLinuxDevice *>(pBoard);

  if (pOmd->pLog) {
    osalinux_free_dma_common_buf(pExternalDevice->m_pPciDev, pOmd->cbSize, (unsigned int) pOmd->pLog, (unsigned int) pOmd->pPhy);
  }

  pOmd->pLog = (void *) NULL;
  pOmd->pPhy = (void *) NULL;
  pOmd->ui64Addr = 0;
  pOmd->pContext = (void *) NULL;

  return true;
}


