/* $Workfile: OsaLinuxDaVinciI2c.c $   $Revision: 5 $    $Modtime: 8/06/10 5:09p $  */
/*******************************************************************************

  Odin TeleSystems Inc.

  OTX Telecom Adapter Driver

  @doc internal

  @module

  Supports communication with I2C devices (specific for Alvis-CSI).

  Device      Address        Comment
  ==========  ============   ==================
  MSP430      0x33  (0x66)   FW#2 and FW#3
  MSP430      0x23  (0x46)   FW#1
  EEprom      0x50
  IOExpander  0x20

  @normal Copyright (C) Odin TeleSystems Inc., 2008-2010, All Rights Reserved

*******************************************************************************/


// Only include with DaVinci
#ifdef DAVINCI


#include <linux/config.h>
#include <linux/init.h>
#include <linux/fs.h>
#include <linux/module.h>
#include <linux/proc_fs.h>
#include <linux/ctype.h>
#include <linux/delay.h>
#include <linux/rtc.h>
#include <linux/bcd.h>
#include <asm/rtc.h>
#include <asm/arch/i2c-client.h>


// DBG is not defined for C files only CPP
#if OSA_DBG
#define DBG 1
#endif
#include "OsaLinuxRtl.h"  // For OSA_DPRINTF
#if OSA_DBG
#undef DBG
#endif

#include "OsaLinuxDaVinciI2c.h"

//
// I2C Message Identifiers
//
#define OD_SET_RTC_PARAMS       0
#define OD_GET_RTC_PARAMS       1
#define OD_SEND_IR_DATA         2   
#define OD_GET_LAST_IR_VALUE    3   // Gets the voltage of the 1.2V and and 1.8V voltage rails in FW#3 and later.
#define OD_GET_INPUT_STATE      4
#define OD_GET_EVENT            5
#define OD_SET_OUTPUT_STATE     6
#define OD_GET_VERSION          7
#define OD_SLEEP_MODE           8
#define OD_GET_TEMPERATURE      9
#define SET_TEMP_LIMIT         10
#define RESET_MAX_TEMP         11

static int g_nMsp430Addr = 0x33; // We will try the old address (0x23) if there is no response at the 0x33 address.

OsaInt32 OsaLinuxDaVinciRtcGetVersion(OsaUint8 *nRtcFwVersion)
{
	char rtcdata [3];
  OsaInt32 rc;

	rtcdata [0] = 2; // Message length
	rtcdata [1] = OD_GET_VERSION; // Command index
	rc = davinci_i2c_write (2, rtcdata, g_nMsp430Addr);
  //OSA_DPRINTF(("davinci_i2c_write(0x%X) rc=%d\n", g_nMsp430Addr, rc));

  if (!rc) {
    udelay (100);
    rc = davinci_i2c_read (3, rtcdata, g_nMsp430Addr);
    //OSA_DPRINTF(("davinci_i2c_read(0x%X) rc=%d\n", g_nMsp430Addr, rc));
  	udelay (100);
  } else {
    
    // Try the old MSP430 Addr
    g_nMsp430Addr = 0x23;
    rtcdata [0] = 2; // Message length
    rtcdata [1] = OD_GET_VERSION; // Command index
    rc = davinci_i2c_write (2, rtcdata, g_nMsp430Addr);
    //OSA_DPRINTF(("davinci_i2c_write(0x%X) rc=%d\n", g_nMsp430Addr, rc));
    if (!rc) {
      udelay (100);
      rc = davinci_i2c_read (3, rtcdata, g_nMsp430Addr);
      //OSA_DPRINTF(("davinci_i2c_read(0x%X) rc=%d\n", g_nMsp430Addr, rc));
      udelay (100);
    }

  }
	
  *nRtcFwVersion = rtcdata[2];

  return rc;
}	

OsaInt32 OsaLinuxDaVinciRtcGetTemperature(OsaInt8 *currentDegC, OsaInt8 *limitDegC, OsaInt8 *maxDegC)
{
	char rtcdata [5];
  OsaInt32 rc;

	rtcdata [0] = 2; // Message length
	rtcdata [1] = OD_GET_TEMPERATURE; // Command index
	rc = davinci_i2c_write (2, rtcdata, g_nMsp430Addr);

  if (!rc) {
    udelay (100);
    rc = davinci_i2c_read (5, rtcdata, g_nMsp430Addr);
  	udelay (100);
  }
	
  *currentDegC = (OsaInt8) rtcdata[2];
  *limitDegC   = (OsaInt8) rtcdata[3];
  *maxDegC     = (OsaInt8) rtcdata[4];

  return rc;
}	

OsaInt32 OsaLinuxDaVinciRtcResetMaxTemp(void)
{
	char rtcdata [5];
  OsaInt32 rc;

	rtcdata [0] = 2; // Message length
	rtcdata [1] = RESET_MAX_TEMP; // Command index
	rc = davinci_i2c_write (2, rtcdata, g_nMsp430Addr);

	udelay (100);
	
  return rc;
}	

OsaInt32 OsaLinuxDaVinciRtcSetTempLimit(OsaInt8 limitDegC)
{
	char rtcdata [5];
  OsaInt32 rc;

	rtcdata [0] = 3; // Message length
	rtcdata [1] = SET_TEMP_LIMIT; // Command index
	rtcdata [2] = limitDegC; // If the temp exceeds this limit we will get interrupts from the RTC
	rc = davinci_i2c_write (3, rtcdata, g_nMsp430Addr);

	udelay (100);
	
  return rc;
}	


OsaInt32 OsaLinuxDaVinciRtcSetOutputState(OsaUint8 p2Val, OsaUint8 p3Val)
{
	char rtcdata [4];
  OsaInt32 rc;

	rtcdata [0] = 4; // Message length
	rtcdata [1] = OD_SET_OUTPUT_STATE; // Command index
	rtcdata [2] = p2Val; // Port2 value
	rtcdata [3] = p3Val; // Port3 value
	rc = davinci_i2c_write (4, rtcdata, g_nMsp430Addr);

	udelay (100);

  return rc;
}	

OsaInt32 OsaLinuxDaVinciRtcGetInputState(OsaUint8 *p2Val, OsaUint8 *p3Val)
{
	char rtcdata [4];
  OsaInt32 rc;

	rtcdata [0] = 2; // Message length
	rtcdata [1] = OD_GET_INPUT_STATE; // Command index
	rc = davinci_i2c_write (2, rtcdata, g_nMsp430Addr);

  if (!rc) {
    udelay (100);
    rc = davinci_i2c_read (4, rtcdata, g_nMsp430Addr);
    udelay (100);
  }

  *p2Val = (OsaUint8) rtcdata[2];
  *p3Val = (OsaUint8) rtcdata[3];

  return rc;
}	

OsaInt32 OsaLinuxDaVinciEepromReadBuf(OsaUint16 nAddress, OsaUint8 *pBuf, OsaInt32 cbBuf)
{
  OsaInt32 rc;
	char data[2];
  
  data[0] = (char) ((nAddress >> 8) & 0xFF);
  data[1] = (char) ((nAddress >> 0) & 0xFF);

  //OSA_DPRINTF(("In OsaLinuxDaVinciEepromReadBuf()\n"));

	rc = davinci_i2c_write (2, data, 0x50);
  if (!rc) {
	  rc = davinci_i2c_read (cbBuf, (char *)pBuf, 0x50);
  }

  return rc;
}

OsaInt32 OsaLinuxDaVinciEepromWritePage(OsaUint16 nAddress, OsaUint8 *pBuf, OsaInt32 cbBuf)
{
  OsaInt32 rc;
	char data[64+2] = { 0x7f, 0, 0xFF}; // 64 for the maximum page size, and 2 for the address
  int  i = 0;

  //OSA_DPRINTF(("In OsaLinuxDaVinciEepromWriteBuf()\n"));

  if (cbBuf > 64) {
    OSA_EPRINTF(("OsaLinuxDaVinciEepromWritePage() Buffer larger than page size. Must be 64 bytes or less.\n"));
    return -1000;
  }

  data[0] = (char) ((nAddress >> 8) & 0xFF);
  data[1] = (char) ((nAddress >> 0) & 0xFF);

  for (i = 0; i < cbBuf; i++) {
    data[i+2] = (char)pBuf[i];
  }

	rc = davinci_i2c_write (cbBuf + 2, data, 0x50);
  udelay (100);
  
  return rc;
}

OsaInt32 OsaLinuxDaVinciI2cRead(OsaInt32 cbBuf, OsaUint8 *pBuf, OsaUint8 nI2cAddress)
{
	return davinci_i2c_read(cbBuf, (char *)pBuf, nI2cAddress);
}

OsaInt32 OsaLinuxDaVinciI2cWrite(OsaInt32 cbBuf, OsaUint8 *pBuf, OsaUint8 nI2cAddress)
{
	return davinci_i2c_write(cbBuf, (char *)pBuf, nI2cAddress);
}


//
// FPGA control implementatino
//
const OsaInt8  DAVINCI_FPGA_I2C_ADDR = 0x54;
const OsaInt32 DAVINCI_FPGA_DELAY    = 100;
//#define DAVINCI_FPGA_I2C_DEBUG_PROCFS


OsaInt32 OsaLinuxFpgaI2CWrite(OsaInt16 subaddr, OsaInt32 data)
{
  subaddr = htons(subaddr);
  OsaInt32 rc = davinci_i2c_write(2, (char*)&subaddr, DAVINCI_FPGA_I2C_ADDR);
  if (!rc) {
    udelay(DAVINCI_FPGA_DELAY);
    data = htonl(data);
    rc = davinci_i2c_write(4, (char*)&data, DAVINCI_FPGA_I2C_ADDR);
  }
  return rc;
}

OsaInt32 OsaLinuxFpgaI2CRead(OsaInt16 subaddr, OsaInt32* data)
{
  subaddr = htons(subaddr);
  OsaInt32 rc = davinci_i2c_write(2, (char*)&subaddr, DAVINCI_FPGA_I2C_ADDR);
  if (!rc) {
    udelay(DAVINCI_FPGA_DELAY);
    rc = davinci_i2c_read(4, (char*)data, DAVINCI_FPGA_I2C_ADDR);
    *data = htonl(*data);
  }
  return rc;
}


#ifdef DAVINCI_FPGA_I2C_DEBUG_PROCFS
/******************************************************************************
  OsaLinuxTdmoeCreateProc
    gives the data to the user
******************************************************************************/
static ssize_t OsaLinuxFpgaI2CReadProc(char *buffer, char **start, off_t offset, int size, int *eof, void *data)
{
  return 0;
}


/******************************************************************************
  OsaLinuxFpgaI2CWrite
    takes data from the user
******************************************************************************/
static ssize_t OsaLinuxFpgaI2CWriteProc(struct file* file, const char* buffer, unsigned long count, void *data)
{
  char m = 0;
  int a = 0;
  int d = 0;
  sscanf(buffer, "%c %x %x", &m, &a, &d);
  if (m == 'w') {
    OsaLinuxFpgaI2CWrite(a, d);
    printk("OsaLinuxFpgaI2C: writing %08x to 0x%04x\n", d, a);
  } else if (m == 'r') {
    OsaLinuxFpgaI2CRead(a, &d);
    printk("OsaLinuxFpgaI2CWrite: result from addr 0x%04x is %08x\n", a, d);
  } else {
    printk("OsaLinuxFpgaI2CWrite: error: command is not recongnized (use \"r addr\" or \"w addr data\")\n");
  }
  return count;
}
#endif


/******************************************************************************
  OsaLinuxTdmoeCreateProc
    creates proc files inside of given directory
******************************************************************************/
extern int OsaLinuxFpgaI2CCreateProc(struct proc_dir_entry* proc)
{
#ifdef DAVINCI_FPGA_I2C_DEBUG_PROCFS
  struct proc_dir_entry* p = create_proc_entry("fpga-i2c", 0644, proc);
  if (p != NULL) {
    p->owner = THIS_MODULE;
    p->read_proc = OsaLinuxFpgaI2CReadProc;
    p->write_proc = OsaLinuxFpgaI2CWriteProc;
  }
#endif
  return 0;
}


/******************************************************************************
  OsaLinuxTdmoeDestroyProc
    destroys proc files inside of given directory
******************************************************************************/
extern int OsaLinuxFpgaI2CDestroyProc(struct proc_dir_entry* proc)
{
#ifdef DAVINCI_FPGA_I2C_DEBUG_PROCFS
  remove_proc_entry("fpga-i2c", proc);
#endif
  return 0;
}



#endif // DAVINCI
