/* $Workfile: Osa.h $   $Revision: 37 $    $Modtime: 6/03/10 5:29p $  */
/*******************************************************************************

  Odin TeleSystems Inc.

  OTX Telecom Adapter Driver

  @doc internal

  @module

  @normal Copyright (C) Odin TeleSystems Inc., 1998-2003, All Rights Reserved

*******************************************************************************/

#ifndef OSA_H
#define OSA_H

// @doc

#include "OsaCppRtl.h"


#define OSA_SUCCESS(nResultCode) (OSA_RES_SUCCESS == (nResultCode))

// Automatically makes string/code translation pairs for inclusion in a translation array
// See also OsaTransStruct and Osa::FindTextForCode()
#define OSA_MK_TRANS_STRUCT(nCode) {#nCode, nCode}   


// forward reference - defined in OSA implementation layer
class OsaAtomicInt;

/******************************************************************************
******************************************************************************/
class OsaAddress64 {
public:
  OsaAddress64(void *p)
    : nQuadPart(0)
  {
    pVoid = p;
  }

  OsaAddress64(OsaInt64 n = 0)
    : nQuadPart(n)
  {}

  OsaAddress64(const OsaAddress64 &that)
    : nQuadPart(nQuadPart)
  {}

  OsaAddress64 &operator=(OsaInt64 n)
  {
    nQuadPart = n;
    return *this;
  }

  OsaAddress64 &operator=(const OsaAddress64 &that)
  {
    nQuadPart = that.nQuadPart;
    return *this;
  }

  operator void*() {
    return pVoid;
  }

  operator OsaUint8*() {
    return (OsaUint8 *) pVoid;
  }

  operator OsaUint16*() {
    return (OsaUint16 *) pVoid;
  }

  operator OsaUint32*() {
    return (OsaUint32 *) pVoid;
  }

  operator OsaInt64() {
    return nQuadPart;
  }

  bool operator!()
  {
    return !nQuadPart;
  }

  union {
    void *pVoid;
    struct {
      OsaUint32 nLow;
      OsaInt32  nHigh;
    } sWord;
    OsaInt64 nQuadPart;
  };
};


/******************************************************************************
  An OsaSpinLock can  be used to synchronize access to data in an
  ISR and SMP safe way. However, remember that the code that executes
  while holding a lock is subject to almost the same limitation as interrupt
  service code with regard operations that can be performed safely.
  
  Note: The implementation is is platform dependent. We want to use
  inlines for these frequently used member function (not virtuals), and we 
  want to enforce the signatures here, but we don't want the implementation
  in this file. Hence the slightly funny declaration of this class. look
  in the OsaXxxRtl.h files for declarations/definitions of OsaSpinLockImpl
  and OsaLocalLockImpl
******************************************************************************/
class OsaSpinLock: public OsaSpinLockImpl {
public:
  void Lock()     { OsaSpinLockImpl::Lock();   } 
  void Unlock()   { OsaSpinLockImpl::Unlock(); } 
  bool IsLocked() { return OsaSpinLockImpl::IsLocked(); } 
};


/******************************************************************************
  A wrapper that helps ensuring a lock is released properly when returning
  at multiple points from a function, without the need to worry about
  unlocking at every return point.

  The OsaSingleLock wrapper is platform independent

******************************************************************************/
class OsaSingleLock {
public:
  OsaSingleLock(OsaSpinLock *pSpinLock)
    : m_pSpinLock(pSpinLock)
  {
    m_pSpinLock->Lock();  
  }
  
  ~OsaSingleLock()
  {
    m_pSpinLock->Unlock();
  }
  
private:
  OsaSpinLock *m_pSpinLock;
};

/******************************************************************************
  deferred procedure callback
******************************************************************************/
typedef void (*POSADPC_CALLBACK)(void *pDpcContext, void *pContext, void *pArg1, void *pArg2);

class OsaDpc: public OsaDpcImpl {
public:
  virtual void Initialize(POSADPC_CALLBACK pFct, void *pContext)
    { OsaDpcImpl::OsaDpcInitialize((void *) pFct, pContext); }
  virtual bool Queue(void *pArg1, void *pArg2)
    { return OsaDpcImpl::OsaDpcInsertQueue(pArg1, pArg2); }
};


/******************************************************************************
  timer
******************************************************************************/

class OsaTimer: public OsaTimerImpl {
public:
  virtual void Initialize(POSADPC_CALLBACK pFct, void *pContext)
    { OsaTimerImpl::OsaTimerInitialize((void *) pFct, pContext); }
  virtual bool TimerSet(OsaUint32 dueTimeInMicroSeconds)
    { return OsaTimerImpl::OsaTimerSet(dueTimeInMicroSeconds); }
};



/******************************************************************************
******************************************************************************/
class OsaEvent: public OsaEventImpl {};
class OsaSemaphore: public OsaSemaphoreImpl {};

/******************************************************************************
******************************************************************************/
typedef void           *OsaExternalDeviceRep;
typedef void           *OsaInternalDeviceRep;

class OsaResourcesS {
public:
  OsaResourcesS() {
    sInterrupt.bHaveInterrupt = false;    
    sInterrupt.nVector = 0;    
    sInterrupt.nLevel = 0;    
    sInterrupt.nProcessorAffinity = 0;
    
    for (int iMemory = 0; iMemory < OSA_BRD_IOMEM_AREAS_MAX; iMemory++) {
      sqMemoryRegArea[iMemory].pStart = 0;
      sqMemoryRegArea[iMemory].cbLength = 0;
      sqMemoryRegArea[iMemory].eAddressSpace = MemorySpace;  // Change to MemorySpaceMapped when mapped
#if defined(_M_AMD64)
      sqMemoryRegArea[iMemory].pVa  = 0;
      sqMemoryRegArea[iMemory].pMdl = 0;
#endif
    }
    
    for (int iPort = 0; iPort < OSA_BRD_IOMEM_AREAS_MAX; iPort++) {
      sqIoPortArea[iPort].pStart = 0;
      sqIoPortArea[iPort].cbLength = 0;
      sqIoPortArea[iPort].eAddressSpace = IoSpace;         // Change to MemorySpaceMapped if mapped
    }
    coMemoryAreas = 0;

  };

  // IoPorts can exist in both IoSpace and MemorySpace. If they exist in memory space
  // they must be mapped before they can be used.
  enum AddressSpaceE {
    MemorySpace = 0,       // Needs mapping before use
    MemorySpaceMapped = 1, // Needs unmapping before exit
    IoSpace     = 2        // Don't need mapping or unmapping
  };

  struct InterruptResourceS {
    bool                bHaveInterrupt;
    OsaUint32           nVector;
    OsaUint32           nLevel;
    bool                bLevelSensitive;
    bool                bMsiInterrupt;
#if defined(_M_AMD64)
    KAFFINITY           nProcessorAffinity;
#else
    OsaUint32           nProcessorAffinity;
#endif
  }; 
  
  struct AddressAreaResourceS {
    OsaAddress64        pStart;
    OsaUint32           cbLength;
    AddressSpaceE       eAddressSpace;
    OsaAddress64        pResource;
#if defined(_M_AMD64)
    OsaAddress64        pVa;        // Use on 64 bit address for physical address
    PMDL                pMdl;
#endif
  };

  InterruptResourceS   sInterrupt;
  OsaUint32            coMemoryAreas;
  AddressAreaResourceS sqMemoryRegArea[OSA_BRD_IOMEM_AREAS_MAX];
  OsaUint32            coPortAreas;
  AddressAreaResourceS sqIoPortArea[OSA_BRD_IOMEM_AREAS_MAX];
};

// @cmember,menum
// Parameter to AddBoard, determines what parameters are used
enum OsaBusTypeE {
  OSA_BUS_PCI,         // @@emem Vendor ID and the Subsystem ID are passed as integers
  OSA_BUS_PCMCIA,      // @@emem A vendor string is passed.
  OSA_BUS_LOCAL,       // @@emem Used if device has direct access to local resource (embeded).
  OSA_BUS_UNKNOWN      // @@emem Represents an unknown bus type, or no bus at all
};

// Commands sent by OSA to the BoardState callback function
enum OsaBoardStateCommandE {
  OSA_BOARD_ACCEPT_NEW_IRPS,                // Accept new IRPs, starting or queueing as appropriated 
  OSA_BOARD_REJECT_NEW_IRPS,                // Do not accept aby new IRPS, return with "device busy"
  OSA_BOARD_HOLD_IRPS,                      // Do not start processing of IRPS, new or queued.
  OSA_BOARD_RESUME_IRPS,                    // Resume processing of IRPs, if none active, start from queue
  OSA_BOARD_CANCEL_QUEUED_IRPS,             // Remove and fail all IRPS in the IRP queues
  OSA_BOARD_DISABLE_IO,                     // Do not allow any I/O writes or reads. Disable all interrupts from board
  OSA_BOARD_ENABLE_IO,                      // Allow I/O writes and reads, allow interrupts from board if appropriate for driver
  OSA_BOARD_CAN_REMOVE,                     // Can the board be removed. (it really should)
  OSA_BOARD_CAN_PAUSE,                      // Can the board be paused (e.g. for rebalancing of resources)
  OSA_BOARD_WAIT_FOR_ACTIVE_IRPS_COMPLETED, // Don't return until all active IRPs have completed
  OSA_BOARD_WAIT_FOR_IRPS_COMPLETED         // Don't return until all IRP queues are empty and all active IRPS have compeleted
};


typedef OsaResultE (*OsaUnloadFnP)          ();
typedef OsaResultE (*OsaAddBoardFnP)        (IN OsaBusTypeE eBusType, IN OsaUint16 nVendorId, IN OsaUint16 nDeviceId, IN OsaUint16 nSubsysId, IN const wchar_t *pwcszVendorString, OUT OsaInternalDeviceRep *pInternalDeviceRep, OUT OsaString *pExternalDeviceName);
typedef OsaResultE (*OsaStartBoardFnP)      (IN OsaInternalDeviceRep internalDeviceRep, IN OsaExternalDeviceRep externalDeviceRep, IN OsaResourcesS *pOsaBoardResources);
typedef OsaResultE (*OsaBoardStateFnP)      (IN OsaInternalDeviceRep internalDeviceRep, IN OsaBoardStateCommandE eCommand);
typedef OsaResultE (*OsaRemoveBoardFnP)     (IN OsaInternalDeviceRep internalDeviceRep);
typedef OsaResultE (*OsaOpenDeviceFnP)      (IN OsaInternalDeviceRep internalDeviceRep, IN OsaIrpP pIrp);  // We pass an IRP since this call may be a blocking/wait call in the future
typedef OsaResultE (*OsaCloseDeviceFnP)     (IN OsaInternalDeviceRep internalDeviceRep, IN OsaIrpP pIrp);  // We pass an IRP since this call may be a blocking/wait call in the future
typedef OsaResultE (*OsaIoControlDeviceFnP) (IN OsaInternalDeviceRep internalDeviceRep, IN OsaIrpP pIrp);
typedef bool       (*OsaSyncCallFnP)        (IN OsaInternalDeviceRep internalDeviceRep, IN OUT void *pContext);

typedef void    (*OsaDeferredBoardNotificationFnP)(IN OsaInternalDeviceRep internalDeviceRep, IN OsaIrpP pIrp, IN void *pContext);
typedef bool    (*OsaBoardNotificationFnP)(IN OsaInternalDeviceRep internalDeviceRep);

// String/code translation pairs for inclusion in a translation array
struct OsaTransStruct {
  const char *pszText;
  OsaUint32  nCode;
};

OsaResultE OsaCreateExternalDevice(
  IN const OsaString &strDeviceName,        // The device name we want to present externally
  IN OsaInternalDeviceRep internalDeviceRep,   // A reference by which the device is known by the internal driver
  OUT OsaExternalDeviceRep *pExternalDeviceRep
  ); 

const OsaString &OsaGetExternalDeviceName(OsaExternalDeviceRep externalDeviceRep);

bool OsaGetDriverCompatibilityMode();
OsaResultE OsaGetCompatibilityVersion(char *sVersion, OsaUint32 bufsiz);
OsaResultE OsaSetCompatibilityVersion(char *sVersion);

OsaResultE OsaGetThor2PciPlusBuildNumberOverride(char *sVersion, OsaUint32 bufsiz);
OsaResultE OsaSetThor2PciPlusBuildNumberOverride(char *sVersion);

OsaResultE OsaGetThor8PciPlusBuildNumberOverride(char *sVersion, OsaUint32 bufsiz);
OsaResultE OsaSetThor8PciPlusBuildNumberOverride(char *sVersion);

OsaUint32  OsaGetAsmHighwaySpeedOverride();
OsaResultE OsaSetAsmHighwaySpeedOverride(char *sHwySpeedInKbps);

OsaResultE OsaSetThor2PciPlusBurstBufferSizeFactor(OsaUint32 nBurstSizeFactor);
OsaResultE OsaSetThor8PciPlusBurstBufferSizeFactor(OsaUint32 nBurstSizeFactor);
OsaResultE OsaSetSleipnir1PciPlusBurstBufferSizeFactor(OsaUint32 nBurstSizeFactor);
OsaResultE OsaSetArniPciPlusBurstBufferSizeFactor(OsaUint32 nBurstSizeFactor);
OsaResultE OsaSetGimle16PciPlusBurstBufferSizeFactor(OsaUint32 nBurstSizeFactor);
OsaResultE OsaSetGimle16PciBasic32BurstBufferSizeFactor(OsaUint32 nBurstSizeFactor);
OsaResultE OsaSetGimle16PciBasic64BurstBufferSizeFactor(OsaUint32 nBurstSizeFactor);
OsaResultE OsaSetThor2ExpressCardBurstBufferSizeFactor(OsaUint32 nBurstSizeFactor);
OsaResultE OsaSetThor4ExpressCardBurstBufferSizeFactor(OsaUint32 nBurstSizeFactor);
OsaResultE OsaSetAlvisPciExpressBurstBufferSizeFactor(OsaUint32 nBurstSizeFactor);
OsaResultE OsaSetGodarPciExpressBurstBufferSizeFactor(OsaUint32 nBurstSizeFactor);
OsaResultE OsaSetDrvEventLogDisplayLevelFactor(OsaUint32 nLogLevelFactor);

OsaUint32 OsaGetThor2PciPlusBurstBufferSizeFactor();
OsaUint32 OsaGetThor8PciPlusBurstBufferSizeFactor();
OsaUint32 OsaGetSleipnir1PciPlusBurstBufferSizeFactor();
OsaUint32 OsaGetArniPciPlusBurstBufferSizeFactor();
OsaUint32 OsaGetGimle16PciPlusBurstBufferSizeFactor();
OsaUint32 OsaGetGimle16PciBasic32BurstBufferSizeFactor();
OsaUint32 OsaGetGimle16PciBasic64BurstBufferSizeFactor();
OsaUint32 OsaGetThor2ExpressCardBurstBufferSizeFactor();
OsaUint32 OsaGetThor4ExpressCardBurstBufferSizeFactor();
OsaUint32 OsaGetAlvisPciExpressBurstBufferSizeFactor();
OsaUint32 OsaGetGodarPciExpressBurstBufferSizeFactor();
OsaUint32 OsaGetDrvEventLogDisplayLevelFactor();


OsaUint32 OsaGetHwLocationAddress(IN OsaExternalDeviceRep externalDeviceRep);
OsaInt32 OsaGetPcmciaSocketNumber(IN OsaExternalDeviceRep externalDeviceRep);
OsaInt32 OsaGetPcmciaControllerNumber(IN OsaExternalDeviceRep externalDeviceRep);

void  OsaDestroyExternalDevice(IN OsaExternalDeviceRep externalDeviceRep);

bool  OsaDoBoardSyncCall(IN OsaExternalDeviceRep pBoard, IN OsaSyncCallFnP pfnSyncCall, IN OUT void *pContext);

void  OsaRequestDeferredBoardNotification(IN OsaExternalDeviceRep pBoard, OsaIrpP pIrp, void *pContext);

void  OsaSetUnloadCallback            (OsaUnloadFnP               );
void  OsaSetAddBoardCallback          (OsaAddBoardFnP             ); 
void  OsaSetStartBoardCallback        (OsaStartBoardFnP           );
void  OsaSetBoardStateCallback        (OsaBoardStateFnP           );
void  OsaSetRemoveBoardCallback       (OsaRemoveBoardFnP          );
              
void  OsaSetOpenDeviceCallback        (OsaOpenDeviceFnP           );
void  OsaSetCloseDeviceCallback       (OsaCloseDeviceFnP          );
void  OsaSetIoControlDeviceCallback   (OsaIoControlDeviceFnP      );

bool  OsaConnectNotificationCallbacks (IN OsaExternalDeviceRep externalDeviceRep, IN OsaBoardNotificationFnP pfnBoardNotificationCallback, IN OsaDeferredBoardNotificationFnP pfnDeferredBoardNotificationCallback, IN OsaResourcesS::InterruptResourceS *pInterruptResource);
void  OsaDisconnectNotificationCallbacks (IN OsaExternalDeviceRep externalDeviceRep);

const char *OsaFindTextForCode(OsaTransStruct sqTextCodeMappings[], OsaUint32 nCode);

// Internal driver part must supply the following function before linking
extern "C" OsaResultE OsaDriverEntry(const wchar_t *pConfigInfoPath);

// OS DMA routines
unsigned long OsaGetDmaMaxMapRegisters(IN OsaExternalDeviceRep pBoard);
bool OsaDmaInitAdapter(IN OsaExternalDeviceRep pBoard, unsigned long ulMaxLength, unsigned long ulMaxFragments);
bool OsaDmaUnInitAdapter(IN OsaExternalDeviceRep pBoard);
bool OsaDmaAllocateAdapter(IN OsaExternalDeviceRep pBoard, IN void * Context);
bool OsaDmaAdapterCheck(IN OsaExternalDeviceRep pBoard);
bool OsaDmaFreeAdapter(IN OsaExternalDeviceRep pBoard, IN void * Context);
bool OsaDmaAllocateCommonBuffer(IN OsaExternalDeviceRep pBoard, IN OUT POSA_MD pOmd);
bool OsaDmaFreeCommonBuffer(IN OsaExternalDeviceRep pBoard, IN OUT POSA_MD pOmd);

#ifdef __cplusplus
extern "C" {
#endif

extern OsaInt32 OsaGetKernelStackAvailable(OsaUint32 *stackAvailable);

#ifdef __cplusplus
}
#endif

#endif
